import os
import re
import sys
import urllib
import webbrowser
from functools import partial
from typing import Dict, Optional

from airunner.components.about.gui.windows.about.about import AboutWindow
from airunner.components.application.gui.widgets.stats.stats_widget import (
    StatsWidget,
)
from airunner.components.application.gui.widgets.status.status_widget import (
    StatusWidget,
)
from airunner.components.application.gui.windows.main.download_model_dialog import (
    show_download_model_dialog,
)
from airunner.components.application.gui.windows.main.nsfw_warning_dialog import (
    show_nsfw_warning_dialog,
)
from airunner.components.art.gui.windows.prompt_browser.prompt_browser import (
    PromptBrowser,
)
from airunner.components.settings.gui.windows.settings.airunner_settings import (
    SettingsWindow,
)
from airunner.components.application.gui.windows.main.model_load_balancer import (
    ModelLoadBalancer,
)
from airunner.components.application.gui.windows.main.worker_manager import (
    WorkerManager,
)
from airunner.components.application.gui.windows.wayland_helper import (
    enable_wayland_window_decorations,
)
import requests
from PIL import Image
from PySide6.QtCore import (
    Slot,
    Signal,
    QProcess,
    QTimer,
    Qt,
)
from PySide6.QtGui import QGuiApplication, QKeySequence, QAction, QCursor
from PySide6.QtWidgets import (
    QApplication,
    QMainWindow,
    QMessageBox,
    QInputDialog,
    QMenu,
)
from PySide6.QtGui import QIcon

from bs4 import BeautifulSoup

from airunner.settings import (
    AIRUNNER_STATUS_ERROR_COLOR,
    AIRUNNER_STATUS_NORMAL_COLOR_LIGHT,
    AIRUNNER_STATUS_NORMAL_COLOR_DARK,
    AIRUNNER_DISCORD_URL,
    AIRUNNER_BUG_REPORT_LINK,
    AIRUNNER_VULNERABILITY_REPORT_LINK,
    AIRUNNER_ART_ENABLED,
)
from airunner.utils.application import create_worker
from airunner.utils.settings import get_qsettings
from airunner.components.llm.managers.llm_request import LLMRequest
from airunner.components.application.data.shortcut_keys import ShortcutKeys
from airunner.components.art.data.image_filter import ImageFilter
from airunner.app_installer import AppInstaller
from airunner.enums import (
    SignalCode,
    GeneratorSection,
    LLMActionType,
    ModelType,
    ModelStatus,
    TemplateName,
)
from airunner.utils.application.mediator_mixin import MediatorMixin
from airunner.utils.application.get_version import get_version
from airunner.utils.widgets import (
    save_splitter_settings,
    load_splitter_settings,
)
from airunner.utils.image import convert_image_to_binary
from airunner.gui.styles.styles_mixin import StylesMixin
from airunner.components.art.gui.windows.filter_window.filter_window import (
    FilterWindow,
)
from airunner.components.application.gui.windows.main.ai_model_mixin import (
    AIModelMixin,
)
from airunner.components.application.gui.windows.main.pipeline_mixin import (
    PipelineMixin,
)
from airunner.components.application.gui.windows.main.settings_mixin import (
    SettingsMixin,
)
from airunner.components.application.gui.windows.main.templates.main_window_ui import (
    Ui_MainWindow,
)
from airunner.components.update.gui.windows.update.update_window import (
    UpdateWindow,
)
from airunner.components.icons.managers.icon_manager import IconManager
from airunner.components.plugins.plugin_loader import PluginLoader


# Utility functions moved from deleted agent.actions
def bash_execute(command: str) -> str:
    """
    Execute a bash command and return the output.

    Args:
        command: The bash command to execute

    Returns:
        The command output or error message
    """
    import subprocess

    try:
        result = subprocess.run(
            command, shell=True, capture_output=True, text=True, timeout=30
        )
        return result.stdout if result.returncode == 0 else result.stderr
    except subprocess.TimeoutExpired:
        return "Command timed out after 30 seconds"
    except Exception as e:
        return f"Error executing command: {str(e)}"


def show_path(path: str) -> None:
    """
    Open the given path in the system file manager.

    Args:
        path: The file or directory path to show
    """
    import subprocess
    import platform
    import os

    if not os.path.exists(path):
        return

    system = platform.system()
    try:
        if system == "Windows":
            os.startfile(path)
        elif system == "Darwin":  # macOS
            subprocess.run(["open", path])
        else:  # Linux
            subprocess.run(["xdg-open", path])
    except Exception:
        pass  # Silently fail if we can't open the path


class MainWindow(
    MediatorMixin,
    SettingsMixin,
    StylesMixin,
    PipelineMixin,
    AIModelMixin,
    QMainWindow,
):
    show_grid_toggled = Signal(bool)
    image_generated = Signal(bool)
    generator_tab_changed_signal = Signal()
    load_image = Signal(str)
    load_image_object = Signal(object)
    loaded = Signal()
    window_opened = Signal()
    ui_class_ = Ui_MainWindow
    _window_title = f"AI Runner"
    icons = [
        ("settings", "actionSettings"),
        ("cpu", "actionToggle_LLM"),
        ("mic", "actionToggle_Speech_to_Text"),
        ("image", "actionToggle_Stable_Diffusion"),
        ("image", "menuArt"),
        ("message-circle", "menuChat"),
        ("refresh-cw", "actionReset_Settings_2"),
        ("x-circle", "actionQuit"),
        ("plus-circle", "artActionNew"),
        ("upload-icon", "actionImport_image"),
        ("download-icon", "actionExport_image_button"),
        ("message-circle", "actionNew_Conversation"),
        ("trash-2", "actionDelete_conversation"),
        ("scissors", "actionCut"),
        ("copy", "actionCopy"),
        ("clipboard", "actionPaste"),
        ("delete", "actionClear_all_prompts"),
        ("settings", "actionSettings"),
        ("book-open", "actionPrompt_Browser"),
        ("folder", "actionBrowse_AI_Runner_Path"),
        ("folder", "actionBrowse_Images_Path_2"),
        ("speaker", "actionToggle_Text_to_Speech"),
        ("image", "menuStable_Diffusion"),
        ("activity", "actionStats"),
        ("zap", "actionRun_setup_wizard_2"),
        ("external-link", "actionBug_report"),
        ("external-link", "actionReport_vulnerability"),
        ("message-square", "actionDiscord"),
        ("download", "actionImport_image"),
        ("upload", "actionExport_image_button"),
        ("codesandbox", "menuWorkflow"),
        ("trash-2", "workflow_actionClear"),
        ("edit", "workflow_actionEdit"),
        ("folder", "workflow_actionOpen"),
        ("pause-circle", "workflow_actionPause"),
        ("play", "workflow_actionRun"),
        ("save", "workflow_actionSave"),
        ("stop-circle", "workflow_actionStop"),
        ("save", "actionSave_As"),
        ("image", "art_editor_button"),
        ("file-text", "document_editor_button"),
        ("calendar", "calendar_button"),
        ("codesandbox", "workflow_editor_button"),
        ("settings", "settings_button"),
        ("message-square", "chat_button"),
        ("home", "home_button"),
        ("radio", "visualizer_button"),
        ("video", "video_button"),
        ("arrow-down-circle", "actionDownload_Model"),
        ("book", "knowledgebase_button"),
        ("file-text", "menuDocuments"),
    ]
    _last_reload_time = 0
    _reload_debounce_seconds = 1.0

    def __init__(self, *args, **kwargs):
        self.quitting = False
        self._state_restored = None
        self.ui = self.ui_class_()
        self.qsettings = get_qsettings()
        self.icon_manager: Optional[IconManager] = None
        self.tab_backup = {}
        self.workflow_tab = None
        self.quitting = False
        self.update_popup = None
        self._document_path = None
        self.prompt = None
        self.negative_prompt = None
        self.image_path = None
        self.token_signal = Signal(str)
        self.input_event_manager = None
        self.tqdm_callback_triggered = False
        self.action = GeneratorSection.TXT2IMG.value
        self.progress_bar_started = False
        self.window = None
        self.canvas = None
        self.models = None
        self.client = None
        self._version = None
        self._latest_version = None
        self.status_error_color = AIRUNNER_STATUS_ERROR_COLOR
        self.status_normal_color_light = AIRUNNER_STATUS_NORMAL_COLOR_LIGHT
        self.status_normal_color_dark = AIRUNNER_STATUS_NORMAL_COLOR_DARK
        self._themes = None
        self.button_clicked_signal = Signal(Dict)
        self.status_widget = None
        self.header_widget_spacer = None
        self.deterministic_window = None
        self.generator = None
        self._generator = None
        self._generator_settings = None
        self.listening = False
        self.initialized = False
        self._model_status = {
            model_type: ModelStatus.UNLOADED for model_type in ModelType
        }
        self.signal_handlers = {
            SignalCode.SD_SAVE_PROMPT_SIGNAL: self.on_save_stablediffusion_prompt_signal,
            SignalCode.QUIT_APPLICATION: self.handle_quit_application_signal,
            SignalCode.WRITE_FILE: self.on_write_file_signal,
            SignalCode.TOGGLE_FULLSCREEN_SIGNAL: self.on_toggle_fullscreen_signal,
            SignalCode.TOGGLE_TTS_SIGNAL: self.on_toggle_tts,
            SignalCode.TOGGLE_LLM_SIGNAL: self.on_toggle_llm,
            SignalCode.APPLICATION_RESET_SETTINGS_SIGNAL: self._action_reset_settings,
            SignalCode.APPLICATION_RESET_PATHS_SIGNAL: self.on_reset_paths_signal,
            SignalCode.MODEL_STATUS_CHANGED_SIGNAL: self.on_model_status_changed_signal,
            SignalCode.KEYBOARD_SHORTCUTS_UPDATED: self.on_keyboard_shortcuts_updated,
            SignalCode.REFRESH_STYLESHEET_SIGNAL: self.on_theme_changed_signal,
            SignalCode.AI_MODELS_SAVE_OR_UPDATE_SIGNAL: self.on_ai_models_save_or_update_signal,
            SignalCode.NAVIGATE_TO_URL: self.on_navigate_to_url,
            SignalCode.MISSING_REQUIRED_MODELS: self.display_missing_models_error,
            SignalCode.ENABLE_WORKFLOWS_TOGGLED: self.on_enable_workflows_toggled,
            SignalCode.RETRANSLATE_UI_SIGNAL: self.on_retranslate_ui_signal,
            SignalCode.APPLICATION_STATUS_ERROR_SIGNAL: self.on_status_error_signal,
        }
        super().__init__()
        self.logger.debug("Starting AI Runnner")
        enable_wayland_window_decorations(self)
        
        # Fix for black background flash during window drag on Linux
        # This ensures the window has a proper background brush during move/resize
        self.setAutoFillBackground(True)
        self.setAttribute(Qt.WidgetAttribute.WA_OpaquePaintEvent, True)
        
        self.update_application_settings(
            sd_enabled=False,
            llm_enabled=False,
            tts_enabled=False,
            stt_enabled=False,
            controlnet_enabled=False,
        )
        self.single_click_timer = QTimer(self)
        self.single_click_timer.setSingleShot(True)
        self.single_click_timer.timeout.connect(self.handle_single_click)
        plugins_path = os.path.join(self.path_settings.base_path, "plugins")
        if plugins_path not in sys.path:
            sys.path.append(plugins_path)
        self._updating_settings = True
        self._updating_settings = False
        self.worker_manager = create_worker(WorkerManager)
        self.model_load_balancer = ModelLoadBalancer(
            self.worker_manager,
            logger=getattr(self, "logger", None),
            api=self.api,
        )
        self.initialize_ui()
        self.last_tray_click_time = 0
        self.settings_window = None

    @property
    def generator_tab_widget(self):
        return self.ui.generator_widget

    @property
    def version(self):
        if self._version is None:
            self._version = get_version()
        return f"v{self._version}"

    @property
    def latest_version(self):
        return self._latest_version

    @latest_version.setter
    def latest_version(self, val):
        self._latest_version = val

    @property
    def document_name(self):
        return "Untitled"

    @property
    def enable_workflows(self) -> bool:
        return self.qsettings.value("enable_workflows") == "true"

    """
    Slot functions
    
    The following functions are defined in and connected to the appropriate
    signals in the corresponding ui file.
    """

    @Slot(bool)
    def on_chat_button_toggled(self, val: bool):
        self._toggle_splitter_section(val, 0, self.ui.main_window_splitter, 50)

    @Slot(bool)
    def on_knowledgebase_button_toggled(self, val: bool):
        self._toggle_splitter_section(val, 2, self.ui.main_window_splitter, 50)

    def on_splitter_changed_sizes(self):
        self.set_chat_button_checked()
        self.set_knowledgebase_button_checked()

    def set_chat_button_checked(self):
        self.ui.chat_button.blockSignals(True)
        self.ui.chat_button.setChecked(
            self.ui.main_window_splitter.sizes()[0] > 0
        )
        self.ui.chat_button.blockSignals(False)

    def set_knowledgebase_button_checked(self):
        panel_id = 2
        self.ui.knowledgebase_button.blockSignals(True)
        self.ui.knowledgebase_button.setChecked(
            self.ui.main_window_splitter.sizes()[panel_id] > 0
        )
        self.ui.knowledgebase_button.blockSignals(False)

    @Slot()
    def on_actionNew_File_triggered(self):
        self.api.document.new_document()

    @Slot()
    def on_actionQuit_triggered(self):
        self.handle_close()

    @Slot()
    def on_actionReset_Settings_2_triggered(self):
        self._action_reset_settings()

    @Slot()
    def on_actionExport_image_button_triggered(self):
        if (
            not self.api
            or not hasattr(self.api, "art")
            or not hasattr(self.api.art, "canvas")
        ):
            self.logger.warning(
                "MainWindow: self.api.art.canvas is missing. Cannot export image."
            )
            return
        self.api.art.canvas.export_image()

    @Slot()
    def on_actionImport_image_triggered(self):
        if (
            not self.api
            or not hasattr(self.api, "art")
            or not hasattr(self.api.art, "canvas")
        ):
            self.logger.warning(
                "MainWindow: self.api.art.canvas is missing. Cannot import image."
            )
            return
        self.api.art.canvas.import_image()

    @Slot()
    def on_artActionNew_triggered(self):
        if (
            not self.api
            or not hasattr(self.api, "art")
            or not hasattr(self.api.art, "canvas")
        ):
            self.logger.warning(
                "MainWindow: self.api.art.canvas is missing. Cannot clear canvas."
            )
            return
        self.api.art.canvas.clear()

    @Slot()
    def on_actionCopy_triggered(self):
        if (
            not self.api
            or not hasattr(self.api, "art")
            or not hasattr(self.api.art, "canvas")
        ):
            self.logger.warning(
                "MainWindow: self.api.art.canvas is missing. Cannot copy image."
            )
            return
        self.api.art.canvas.copy_image()

    @Slot()
    def on_actionClear_all_prompts_triggered(self):
        self.clear_all_prompts()

    @Slot()
    def on_actionBrowse_AI_Runner_Path_triggered(self):
        # Note: show_path functionality removed with old agent system
        # path = self.path_settings.base_path
        # if path == "":
        #     path = AIRUNNER_BASE_PATH
        # TODO: Implement file browser opening if needed
        pass

    @Slot()
    def on_actionDownload_Model_triggered(self):
        show_download_model_dialog(
            self, self.path_settings, self.application_settings
        )

    @Slot()
    def action_show_model_path_txt2img(self):
        self.show_settings_path("txt2img_model_path")

    @Slot()
    def action_show_model_path_inpaint(self):
        self.show_settings_path("inpaint_model_path")

    @Slot()
    def action_show_model_path_embeddings(self):
        self.show_settings_path("embeddings_model_path")

    @Slot()
    def action_show_model_path_lora(self):
        self.show_settings_path("lora_model_path")

    @Slot()
    def action_show_llm(self):
        pass

    @Slot()
    def on_actionReport_vulnerability_triggered(self):
        webbrowser.open(AIRUNNER_VULNERABILITY_REPORT_LINK)

    @Slot()
    def on_actionBug_report_triggered(self):
        webbrowser.open(AIRUNNER_BUG_REPORT_LINK)

    @Slot()
    def on_actionDiscord_triggered(self):
        if AIRUNNER_DISCORD_URL:
            webbrowser.open(AIRUNNER_DISCORD_URL)

    @Slot(bool)
    def action_outpaint_toggled(self, val: bool):
        self.update_outpaint_settings(enabled=val)

    @Slot()
    def action_outpaint_export(self):
        pass

    @Slot()
    def action_outpaint_import(self):
        pass

    @Slot()
    def on_actionRun_setup_wizard_2_triggered(self):
        self.show_setup_wizard()

    @Slot(bool)
    def on_actionToggle_Stable_Diffusion_toggled(self, val: bool):
        self._update_action_button(
            ModelType.SD,
            self.ui.actionToggle_Stable_Diffusion,
            val,
            SignalCode.SD_LOAD_SIGNAL,
            SignalCode.SD_UNLOAD_SIGNAL,
            "sd_enabled",
        )

    @Slot()
    def on_actionSettings_triggered(self):
        self._show_settings_window()

    @Slot()
    def on_actionBrowse_Images_Path_2_triggered(self):
        self.show_settings_path("image_path")

    @Slot()
    def on_actionPrompt_Browser_triggered(self):
        PromptBrowser()

    @Slot()
    def on_actionStats_triggered(self):
        widget = StatsWidget()
        # display in a window
        widget.show()

    @Slot(bool)
    def on_actionToggle_LLM_toggled(self, val: bool):
        self.on_toggle_llm(val=val)

    @Slot(bool)
    def on_actionToggle_Speech_to_Text_toggled(self, val: bool):
        if self._model_status[ModelType.STT] is ModelStatus.LOADING:
            val = not val
        self._update_action_button(
            ModelType.STT,
            self.ui.actionToggle_Speech_to_Text,
            val,
            SignalCode.STT_LOAD_SIGNAL,
            SignalCode.STT_UNLOAD_SIGNAL,
            "stt_enabled",
        )
        QApplication.processEvents()
        self.update_application_settings(stt_enabled=val)

    @Slot()
    def on_workflow_actionClear_triggered(self):
        self.ui.graph.clear_graph()

    @Slot()
    def on_workflow_actionRun_triggered(self):
        self.ui.graph.run_workflow()

    @Slot()
    def on_workflow_actionEdit_triggered(self):
        self.ui.graph.edit_workflow()

    @Slot()
    def on_workflow_actionPause_triggered(self):
        self.ui.graph.pause_workflow()

    @Slot()
    def on_workflow_actionSave_triggered(self):
        self.ui.graph.save_workflow()

    @Slot()
    def on_workflow_actionStop_triggered(self):
        self.ui.graph.stop_workflow()

    @Slot()
    def on_workflow_actionOpen_triggered(self):
        self.ui.graph.load_workflow()

    @Slot(bool)
    def on_actionToggle_Text_to_Speech_toggled(self, val: bool):
        self.on_toggle_tts(val=val)

    @Slot(bool)
    def on_actionSafety_Checker_toggled(self, val: bool):
        """Handle safety checker toggle action."""
        # If disabling the safety checker, show warning dialog
        if not val:
            # Check if we should show the warning (user hasn't chosen to hide it)
            settings = get_qsettings()
            show_warning = settings.value(
                "nsfw_warning/show_again", True, type=bool
            )

            if show_warning:
                confirmed, do_not_show_again = show_nsfw_warning_dialog(
                    self, show_again_default=show_warning
                )

                # Save the "do not show again" preference
                if do_not_show_again:
                    settings.setValue("nsfw_warning/show_again", False)

                # If user cancelled, revert the checkbox
                if not confirmed:
                    if hasattr(self.ui, "actionSafety_Checker"):
                        self.ui.actionSafety_Checker.blockSignals(True)
                        self.ui.actionSafety_Checker.setChecked(True)
                        self.ui.actionSafety_Checker.blockSignals(False)
                    return

        # Update the setting
        self.update_application_settings(nsfw_filter=val)
        self.set_nsfw_filter_tooltip()

        # Emit signal to load or unload the safety checker worker
        from airunner.enums import SignalCode

        if val:
            self.emit_signal(SignalCode.SAFETY_CHECKER_LOAD_SIGNAL, {})
        else:
            self.emit_signal(SignalCode.SAFETY_CHECKER_UNLOAD_SIGNAL, {})

    def set_nsfw_filter_tooltip(self):
        """Update the safety checker button tooltip based on current state."""
        nsfw_filter = self.application_settings.nsfw_filter
        if hasattr(self.ui, "actionSafety_Checker"):
            self.ui.actionSafety_Checker.setToolTip(
                f"Click to {'enable' if not nsfw_filter else 'disable'} NSFW filter"
            )

    def _add_legal_menu_items(self):
        """Add Terms of Service, Privacy Policy, and Age Agreement menu items to Help menu."""
        from PySide6.QtGui import QAction
        
        # Add separator before legal items
        self.ui.menuAbout.addSeparator()
        
        # Age Agreement action
        self.actionAgeAgreement = QAction("Age Restriction Policy", self)
        self.actionAgeAgreement.triggered.connect(self._show_age_agreement)
        self.ui.menuAbout.addAction(self.actionAgeAgreement)
        
        # Terms of Service action
        self.actionTermsOfService = QAction("Terms of Service", self)
        self.actionTermsOfService.triggered.connect(self._show_terms_of_service)
        self.ui.menuAbout.addAction(self.actionTermsOfService)
        
        # Privacy Policy action
        self.actionPrivacyPolicy = QAction("Privacy Policy", self)
        self.actionPrivacyPolicy.triggered.connect(self._show_privacy_policy)
        self.ui.menuAbout.addAction(self.actionPrivacyPolicy)

    def _add_download_models_menu_item(self):
        """Add Download Models action to Tools menu."""
        from PySide6.QtGui import QAction
        
        # Add separator before download models
        self.ui.menuTools.addSeparator()
        
        # Download Models action
        self.actionDownloadModels = QAction("Download Models...", self)
        self.actionDownloadModels.setToolTip(
            "Download pre-configured models from HuggingFace"
        )
        self.actionDownloadModels.triggered.connect(self._show_download_models_dialog)
        self.ui.menuTools.addAction(self.actionDownloadModels)
        
        # Privacy Settings action
        self.actionPrivacySettings = QAction("Privacy Settings...", self)
        self.actionPrivacySettings.setToolTip(
            "Manage external service connections and privacy options"
        )
        self.actionPrivacySettings.triggered.connect(self._show_privacy_settings)
        self.ui.menuTools.addAction(self.actionPrivacySettings)
    
    def _show_privacy_settings(self):
        """Show the Privacy Settings dialog."""
        from airunner.components.application.gui.dialogs.privacy_consent_dialog import (
            PrivacyConsentDialog,
        )
        dialog = PrivacyConsentDialog(self)
        dialog.exec()

    def _show_download_models_dialog(self):
        """Show the Download Models dialog."""
        from airunner.components.application.gui.dialogs.download_models_dialog import (
            show_download_models_dialog,
        )
        show_download_models_dialog(self)
    
    def _show_age_agreement(self):
        """Show Age Agreement dialog."""
        from airunner.components.application.gui.dialogs.legal_document_dialog import (
            LegalDocumentDialog,
        )
        dialog = LegalDocumentDialog(
            self,
            title="Age Restriction Policy",
            document_type="age"
        )
        dialog.exec()
    
    def _show_terms_of_service(self):
        """Show Terms of Service dialog."""
        from airunner.components.application.gui.dialogs.legal_document_dialog import (
            LegalDocumentDialog,
        )
        dialog = LegalDocumentDialog(
            self,
            title="Terms of Service",
            document_type="terms"
        )
        dialog.exec()
    
    def _show_privacy_policy(self):
        """Show Privacy Policy dialog."""
        from airunner.components.application.gui.dialogs.legal_document_dialog import (
            LegalDocumentDialog,
        )
        dialog = LegalDocumentDialog(
            self,
            title="Privacy Policy",
            document_type="privacy"
        )
        dialog.exec()

    @Slot()
    def on_actionAbout_triggered(self):
        AboutWindow()

    @Slot()
    def on_actionNew_Conversation_triggered(self):
        if not self.api or not hasattr(self.api, "llm"):
            self.logger.warning(
                "MainWindow: self.api.llm is missing. Cannot clear LLM history."
            )
            return
        self.api.llm.clear_history()

    @Slot()
    def on_actionDelete_conversation_triggered(self):
        if not self.api or not hasattr(self.api, "llm"):
            self.logger.warning(
                "MainWindow: self.api.llm is missing. Cannot delete conversation."
            )
            return
        current_conversation = self.llm_generator_settings.current_conversation
        self.api.llm.converation_deleted(current_conversation.id)

    def _set_tab_index(self, tab_widget):
        index = self.ui.center_tab_container.indexOf(tab_widget)
        # Get current window settings first
        current_settings = self.window_settings
        # Update only the active_main_tab_index, keeping other values unchanged
        updated_settings = {
            "is_maximized": current_settings.is_maximized,
            "is_fullscreen": current_settings.is_fullscreen,
            "width": current_settings.width,
            "height": current_settings.height,
            "x_pos": current_settings.x_pos,
            "y_pos": current_settings.y_pos,
            "active_main_tab_index": index,
        }
        self.window_settings = updated_settings
        self.ui.center_tab_container.setCurrentIndex(index)

    @property
    def buttons(self) -> Dict:
        return {
            "home_button": self.ui.home_tab,
            "art_editor_button": self.ui.art_tab,
            "document_editor_button": self.ui.document_editor_tab,
            "calendar_button": self.ui.calendar_tab,
            "workflow_editor_button": self.ui.agent_workflow_tab,
            "visualizer_button": self.ui.visualizer_tab,
            "video_button": self.ui.video_tab,
        }

    def _restore_tab(self):
        """
        Restore the last active tab based on the saved index in QSettings.
        """
        # Read the saved tab index directly from QSettings to avoid getter issues
        # from airunner.utils.settings import get_qsettings

        # settings = get_qsettings()
        # settings.beginGroup("window_settings")
        # saved_index = settings.value("active_main_tab_index", 0, type=int)
        # settings.endGroup()

        saved_index = self.window_settings.active_main_tab_index

        buttons = {
            0: "home_button",
            1: "art_editor_button",
            2: "workflow_editor_button",
            3: "document_editor_button",
            4: "calendar_button",
        }

        if saved_index in buttons:
            self.ui.center_tab_container.setCurrentIndex(saved_index)
            button_name = buttons[saved_index]
            getattr(self.ui, button_name).setChecked(True)

    def _set_current_button_and_tab(self, button_name: str):
        """
        Set the current button and tab based on the button name and tab widget.
        Also emits SECTION_CHANGED signal to notify LLM of UI context change.
        """
        buttons = self.buttons.keys()
        for btn in buttons:
            getattr(self.ui, btn).blockSignals(True)
            getattr(self.ui, btn).setChecked(btn == button_name)
            getattr(self.ui, btn).blockSignals(False)
        tab_widget = self.buttons[button_name]
        self._set_tab_index(tab_widget)
        
        # Emit section changed signal with button name for LLM context
        self.emit_signal(SignalCode.SECTION_CHANGED, {"section": button_name})

    @Slot(bool)
    def on_home_button_toggled(self, active: bool):
        self._set_current_button_and_tab("home_button")

    @Slot(bool)
    def on_visualizer_button_toggled(self, val: bool):
        self._set_current_button_and_tab("visualizer_button")

    @Slot(bool)
    def on_art_editor_button_toggled(self, val: bool):
        self._set_current_button_and_tab("art_editor_button")

    @Slot(bool)
    def on_document_editor_button_toggled(self, val: bool):
        self._set_current_button_and_tab("document_editor_button")

    @Slot(bool)
    def on_calendar_button_toggled(self, val: bool):
        self._set_current_button_and_tab("calendar_button")

    @Slot(bool)
    def on_workflow_editor_button_toggled(self, val: bool):
        self._set_current_button_and_tab("workflow_editor_button")

    @Slot(bool)
    def on_video_button_toggled(self, val: bool):
        self._set_current_button_and_tab("video_button")

    @Slot(bool)
    def on_settings_button_clicked(self, val: bool):
        self._show_settings_window()

    def _show_settings_window(self):
        if self.settings_window is None:
            self.settings_window = SettingsWindow(
                prevent_always_on_top=False, exec=False
            )
            self.settings_window.show()
        elif not self.settings_window.isVisible():
            self.settings_window.show()
        self.settings_window.raise_()

    def _action_reset_settings(self):
        reply = QMessageBox.question(
            self,
            "Reset Settings",
            "Are you sure you want to reset all settings to their default values?",
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No,
        )

        if reply == QMessageBox.Yes:
            self.reset_settings()
            self.restart()

    """
    End slot functions
    """

    @staticmethod
    def download_url(url, save_path):
        response = requests.get(url)
        soup = BeautifulSoup(response.content, "html.parser")
        title = soup.title.string if soup.title else url
        # Truncate title to 10 words
        title_words = title.split()[:10]
        filename = "_".join(title_words) + ".html"
        # Replace any characters in filename that are not alphanumerics, underscores, or hyphens
        filename = re.sub(r"[^\w\-_]", "_", filename)
        save_path = os.path.join(save_path, filename)
        with open(save_path, "wb") as file:
            file.write(response.content)
        return filename

    @staticmethod
    def download_pdf(url, save_path):
        response = requests.get(url)
        filename = url.split("/")[-1]  # Get the filename from the URL
        save_path = os.path.join(save_path, filename)
        with open(save_path, "wb") as file:
            file.write(response.content)
        return filename

    def on_navigate_to_url(self, _data: Dict = None):
        url, ok = QInputDialog.getText(self, "Browse Web", "Enter your URL:")
        if ok:
            try:
                result = urllib.parse.urlparse(url)
                is_url = all([result.scheme, result.netloc])
            except ValueError:
                is_url = False

            # If the input is a URL, download it
            if is_url:
                if url.lower().endswith(".pdf"):
                    # Handle PDF file
                    filepath = os.path.expanduser(self.path_settings.pdf_path)
                    filename = self.download_pdf(url, filepath)
                else:
                    # Handle URL
                    filepath = os.path.expanduser(
                        self.path_settings.webpages_path
                    )
                    filename = self.download_url(url, filepath)
            elif os.path.isfile(url):
                filepath = os.path.dirname(url)
                filename = os.path.basename(url)
            else:
                self.logger.error(f"Invalid URL or file path")
                return

            # Update target files to use only the file that was downloaded or navigated to
            # and update the index.
            self.update_chatbot(
                "target_files", [os.path.join(filepath, filename)]
            )
            if not self.api or not hasattr(self.api, "llm"):
                self.logger.warning(
                    "MainWindow: self.api.llm is missing. Cannot reload RAG."
                )
                return
            self.api.llm.reload_rag(self.chatbot.target_files)
            if not self.api or not hasattr(self.api, "llm"):
                self.logger.warning(
                    "MainWindow: self.api.llm is missing. Cannot send LLM request."
                )
                return
            self.api.llm.send_request(
                action=LLMActionType.RAG,
                prompt="Summarize the text and provide a synopsis of the content. Be concise and informative.",
                llm_request=LLMRequest.from_default(),
            )

    def on_reset_paths_signal(self):
        self.reset_path_settings()

    def restart(self):
        # Save the current state
        self.save_state()

        # Close the main window
        self.close()

        # Start a new instance of the application
        QProcess.startDetached(sys.executable, sys.argv)

    def toggle_window_visibility(self):
        """Toggle window visibility and update the menu text."""
        if self.isVisible():
            self.hide()
            self.toggle_visibility_action.setText("Show Window")
        else:
            self.showNormal()
            self.activateWindow()  # Ensure window gets focus when showing
            self.toggle_visibility_action.setText("Hide Window")

        # Update the tray menu so it reflects the current state immediately
        self.tray_icon.setContextMenu(self.tray_menu)

    def handle_single_click(self):
        """Handle single-click on the tray icon."""
        # Create a dropdown menu
        menu = QMenu()

        # Dynamically set the text based on current window visibility
        show_hide_text = "Hide Window" if self.isVisible() else "Show Window"
        show_action = QAction(show_hide_text, self)
        quit_action = QAction("Quit", self)

        # Connect actions to their respective slots
        show_action.triggered.connect(self.toggle_window_visibility)
        quit_action.triggered.connect(QApplication.quit)

        # Add actions to the menu
        menu.addAction(show_action)
        menu.addAction(quit_action)

        # Display the menu under the tray icon
        menu.exec(QCursor.pos())

    def handle_double_click(self):
        """Handle double-click on the tray icon."""
        self.toggle_window_visibility()  # Use toggle instead of just showing

    @staticmethod
    def on_write_file_signal(data: Dict):
        """
        Writes data to a file.
        :param data: Dict
        :return: None
        """
        args = data["args"]
        if len(args) == 1:
            message = args[0]
        else:
            message = args[1]
        with open("output.txt", "w") as f:
            f.write(message)

    def on_theme_changed_signal(self, data: Dict):
        template = data.get("template", TemplateName.DARK)
        self.set_stylesheet(
            template=template,
        )

    def initialize_ui(self):
        self.logger.debug("Loading UI")

        self.ui.setupUi(self)

        # Disable innactive features
        self.ui.video_button.hide()
        self.ui.calendar_button.hide()
        
        # Add legal document menu items to Help menu
        self._add_legal_menu_items()
        
        # Add Download Models menu item to Tools menu
        self._add_download_models_menu_item()

        self.icon_manager = IconManager(self.icons, self.ui)

        if not AIRUNNER_ART_ENABLED:
            self._disable_aiart_gui_elements()

        # Restore last active tab index from QSettings
        self._restore_tab()

        # Store active tab index on tab change
        self.ui.center_tab_container.currentChanged.connect(
            self._store_active_tab_index
        )

        self.set_stylesheet()
        self.icon_manager.set_icons()
        self.restore_state()
        # Configure default splitter sizes to maximize the canvas area (index 1)
        default_splitter_config = {
            "main_window_splitter": {
                "index_to_maximize": 1,
                "min_other_size": 50,
            }
        }
        load_splitter_settings(
            self.ui,
            ["main_window_splitter"],
            default_maximize_config=default_splitter_config,
            namespace="MainWindow",
        )

        self.status_widget = StatusWidget()
        self.statusBar().addPermanentWidget(self.status_widget)
        if not self.api:
            self.logger.warning(
                "MainWindow: self.api is missing. Cannot clear status message."
            )
            return
        self.api.clear_status_message()
        self.initialize_widget_elements()
        self.last_tray_click_time = 0
        self.settings_window = None
        self.hide_center_tab_header()
        self._load_plugins()

        self.ui.main_window_splitter.splitterMoved.connect(
            self.on_splitter_changed_sizes
        )
        self.set_chat_button_checked()
        self.set_knowledgebase_button_checked()

        # Keyboard shortcut: Ctrl+N -> placeholder handler using QAction
        try:
            action_new_shortcut = QAction(self)
            action_new_shortcut.setShortcut(QKeySequence("Ctrl+N"))
            action_new_shortcut.triggered.connect(self._on_ctrl_n_pressed)
            # add to window so the shortcut is active
            self.addAction(action_new_shortcut)
            self._action_new_shortcut = action_new_shortcut
        except Exception:
            # Fail silently if QAction creation fails
            try:
                self.logger.debug("Could not create Ctrl+N QAction shortcut")
            except Exception:
                pass

    def _disable_aiart_gui_elements(self):
        self.ui.center_widget.hide()
        self.ui.menuImage.hide()
        self.ui.menuFilters.hide()
        self.ui.menuStable_Diffusion.hide()
        self.ui.menuArt.hide()

        self.ui.center_widget.deleteLater()
        self.ui.actionToggle_Stable_Diffusion.deleteLater()
        self.ui.menuImage.deleteLater()
        self.ui.menuFilters.deleteLater()
        self.ui.menuStable_Diffusion.deleteLater()
        self.ui.actionBrowse_AI_Runner_Path.deleteLater()
        self.ui.actionBrowse_Images_Path_2.deleteLater()
        self.ui.actionArt.deleteLater()
        self.ui.menuArt.deleteLater()
        self.ui.actionCut.deleteLater()
        self.ui.actionCopy.deleteLater()
        self.ui.actionPaste.deleteLater()
        self.ui.actionPrompt_Browser.deleteLater()

    def _load_plugins(self):
        base_path = self.path_settings.base_path
        path = os.path.join(base_path, "plugins")
        plugin_loader = PluginLoader(plugin_dir=path)
        plugins = plugin_loader.load_plugins()

        if len(plugins) > 0:
            self.logger.info("Loading plugins")
            for plugin in plugins:
                if hasattr(plugin, "get_widget"):
                    widget = plugin.get_widget()
                    self.ui.center_tab_container.addTab(widget, plugin.name)

    def initialize_widget_elements(self):
        for item in (
            (self.ui.actionToggle_LLM, self.application_settings.llm_enabled),
            (
                self.ui.actionToggle_Text_to_Speech,
                self.application_settings.tts_enabled,
            ),
            (
                self.ui.actionToggle_Speech_to_Text,
                self.application_settings.stt_enabled,
            ),
            (
                self.ui.actionToggle_Stable_Diffusion,
                self.application_settings.sd_enabled,
            ),
        ):
            item[0].blockSignals(True)
            item[0].setChecked(item[1] or False)
            item[0].blockSignals(False)

        # Initialize safety checker action if it exists
        if hasattr(self.ui, "actionSafety_Checker"):
            self.ui.actionSafety_Checker.blockSignals(True)
            self.ui.actionSafety_Checker.setChecked(
                self.application_settings.nsfw_filter
            )
            self.ui.actionSafety_Checker.blockSignals(False)
            self.set_nsfw_filter_tooltip()

        self.initialized = True

    def keyPressEvent(self, event):
        super().keyPressEvent(event)
        for v in self.shortcut_keys:
            if v.key == event.key():
                for signal in SignalCode:
                    if signal.value == v.signal:
                        self.emit_signal(signal)
                        break

    def key_text(self, key_name):
        for shortcutkey in self.shortcut_keys:
            if shortcutkey.name == key_name:
                return shortcutkey.text
        return ""

    @property
    def current_active_tab_index(self) -> int:
        """Get the current active main tab index from QSettings."""
        self.qsettings.beginGroup("window_settings")
        index = self.qsettings.value("active_main_tab_index", 0, type=int)
        self.qsettings.endGroup()
        return index

    @Slot()
    def _on_ctrl_n_pressed(self):
        """Placeholder handler for Ctrl+N keyboard shortcut.

        Currently this is a stub that logs the event and updates the status
        bar. Replace with desired behavior (e.g. create new document) as needed.
        """
        current_active_tab_index = self.current_active_tab_index
        if current_active_tab_index == 1:
            self.api.art.canvas.new_document()
        elif current_active_tab_index == 2:
            self.api.nodegraph.new_document()
        elif current_active_tab_index == 3:
            self.api.document.new_document()
        self.api.llm.clear_history()

    def on_save_stablediffusion_prompt_signal(self, data: Dict):
        self.create_saved_prompt(
            {
                "prompt": data["prompt"],
                "negative_prompt": data["negative_prompt"],
                "secondary_prompt": data["secondary_prompt"],
                "secondary_negative_prompt": data["secondary_negative_prompt"],
            }
        )

    def set_path_settings(self, key, val):
        self.update_path_settings(**{key: val})

    def closeEvent(self, event):
        event.ignore()
        self.handle_close()

    def handle_close(self):
        """Override close to minimize to tray instead of exiting."""
        self.quit()

    def quit(self):
        if self.quitting:
            return
        self.logger.debug("Quitting")
        self.save_state()
        self.quitting = True
        if not self.api:
            self.logger.warning(
                "MainWindow: self.api is missing. Cannot quit application."
            )
            return
        self.emit_signal(SignalCode.SAVE_STATE, {})
        self.emit_signal(SignalCode.QUIT_APPLICATION, {})

    def handle_quit_application_signal(self):
        self.hide()
        QApplication.quit()
        self.close()

    def show_settings_path(self, name, default_path=None):
        # Note: show_path functionality removed with old agent system
        # path = getattr(self.path_settings, name)
        # TODO: Implement file browser opening if needed
        pass

    def on_toggle_fullscreen_signal(self):
        if self.isFullScreen():
            self.showNormal()
        else:
            self.showFullScreen()

    def on_toggle_llm(self, data: Dict = None, val=None):
        if val is None:
            val = data.get(
                "enabled", not self.application_settings.llm_enabled
            )
        self._update_action_button(
            ModelType.LLM,
            self.ui.actionToggle_LLM,
            val,
            SignalCode.LLM_LOAD_SIGNAL,
            SignalCode.LLM_UNLOAD_SIGNAL,
            "llm_enabled",
            data,
        )

    def on_toggle_tts(self, data: Dict = None, val=None):
        if val is None:
            val = data.get(
                "enabled", not self.application_settings.tts_enabled
            )
        self._update_action_button(
            ModelType.TTS,
            self.ui.actionToggle_Text_to_Speech,
            val,
            SignalCode.TTS_ENABLE_SIGNAL,
            SignalCode.TTS_DISABLE_SIGNAL,
            "tts_enabled",
            data,
        )

    def _update_action_button(
        self,
        model_type,
        element,
        val: bool,
        load_signal: SignalCode,
        unload_signal: SignalCode,
        application_setting: str = None,
        data: Dict = None,
    ):
        if self._model_status[model_type] is ModelStatus.LOADING:
            val = not val
        element.blockSignals(True)
        element.setChecked(val)
        element.blockSignals(False)
        QApplication.processEvents()
        if application_setting:
            settings_data = {}
            settings_data[application_setting] = val
            self.update_application_settings(**settings_data)
        if self._model_status[model_type] is not ModelStatus.LOADING:
            if val:
                self.emit_signal(load_signal, data)
            else:
                self.emit_signal(unload_signal, data)

    def save_state(self):
        self.logger.debug("Saving window state")

        self.qsettings.beginGroup("window_settings")
        is_maximized = self.isMaximized()
        is_fullscreen = self.isFullScreen()
        self.logger.debug(
            f"Saving state - maximized: {is_maximized}, fullscreen: {is_fullscreen}"
        )
        self.qsettings.setValue("is_maximized", is_maximized)
        self.qsettings.setValue("is_fullscreen", is_fullscreen)

        # Always save current geometry and position
        width = self.width()
        height = self.height()
        x_pos = self.pos().x()
        y_pos = self.pos().y()
        self.logger.debug(
            f"Saving geometry - width: {width}, height: {height}, x: {x_pos}, y: {y_pos}"
        )
        self.qsettings.setValue("width", width)
        self.qsettings.setValue("height", height)
        self.qsettings.setValue("x_pos", x_pos)
        self.qsettings.setValue("y_pos", y_pos)

        # Save which screen the window is on
        try:
            screen = self.screen()
            if screen:
                screen_name = screen.name()
                self.qsettings.setValue("screen_name", screen_name)
                self.logger.debug(f"Saving screen: {screen_name}")
        except Exception:
            self.logger.exception("Failed to save screen information")

        self.qsettings.setValue(
            "active_main_tab_index",
            self.ui.center_tab_container.currentIndex(),
        )
        self.qsettings.endGroup()
        
        # Ensure settings are written to disk before exit
        self.qsettings.sync()
        
        save_splitter_settings(self.ui, ["main_window_splitter"], "MainWindow")

        # Save canvas offset for all canvas views
        try:
            if hasattr(self.ui, "image_canvas"):
                self.ui.image_canvas.save_canvas_offset()
            if hasattr(self.ui, "brush_canvas"):
                self.ui.brush_canvas.save_canvas_offset()
        except Exception:
            self.logger.exception("Failed to save canvas offset")

    def restore_state(self):
        """
        Restore the window based on the previous state using QSettings.
        """
        self.qsettings.beginGroup("window_settings")
        is_maximized = self.qsettings.value("is_maximized", False, type=bool)
        is_fullscreen = self.qsettings.value("is_fullscreen", False, type=bool)
        width = self.qsettings.value("width", 1024, type=int)
        height = self.qsettings.value("height", 768, type=int)
        x_pos = self.qsettings.value("x_pos", 100, type=int)
        y_pos = self.qsettings.value("y_pos", 100, type=int)
        screen_name = self.qsettings.value("screen_name", None, type=str)
        self.qsettings.endGroup()

        self.logger.debug(
            f"Restoring state - maximized: {is_maximized}, fullscreen: {is_fullscreen}, "
            f"geometry: {width}x{height} at ({x_pos}, {y_pos}), screen: {screen_name}"
        )

        # Try to restore to the same screen
        target_screen = None
        if screen_name:
            try:
                from PySide6.QtGui import QGuiApplication

                for screen in QGuiApplication.screens():
                    if screen.name() == screen_name:
                        target_screen = screen
                        self.logger.debug(
                            f"Found target screen: {screen_name}"
                        )
                        break
                if not target_screen:
                    self.logger.warning(
                        f"Could not find screen: {screen_name}, using primary"
                    )
            except Exception:
                self.logger.exception("Error finding target screen")

        # If we found a target screen, set it BEFORE positioning/showing the window
        if target_screen:
            try:
                self.create()  # Ensure native window is created
                if self.windowHandle():
                    self.windowHandle().setScreen(target_screen)

                    # Calculate position relative to the target screen
                    screen_geometry = target_screen.geometry()
                    if is_maximized or is_fullscreen:
                        # For maximized/fullscreen, move to screen's top-left
                        self.move(screen_geometry.x(), screen_geometry.y())
                    else:
                        # For normal windows, ensure saved position is on the target screen
                        # If saved position is outside target screen, center it
                        if not screen_geometry.contains(x_pos, y_pos):
                            centered_x = (
                                screen_geometry.x()
                                + (screen_geometry.width() - width) // 2
                            )
                            centered_y = (
                                screen_geometry.y()
                                + (screen_geometry.height() - height) // 2
                            )
                            self.move(centered_x, centered_y)
                        else:
                            self.move(x_pos, y_pos)
            except Exception:
                self.logger.exception("Error setting window screen")
        else:
            # No target screen found, use saved position
            self.move(x_pos, y_pos)

        # Set size
        self.resize(width, height)
        self.setMinimumSize(512, 512)

        # Force the screen again right before showing (Qt sometimes resets it)
        if target_screen:
            try:
                if self.windowHandle():
                    self.windowHandle().setScreen(target_screen)
            except Exception:
                self.logger.exception(
                    "Error re-setting window screen before show"
                )

        # Mark that state has been restored BEFORE showing to prevent _initialize_window from overriding
        self._state_restored = True

        # Now apply special states if needed
        if is_maximized:
            self.logger.info("Applying maximized state")
            # Move to screen one more time right before maximizing
            if target_screen:
                screen_geometry = target_screen.geometry()
                self.move(screen_geometry.x(), screen_geometry.y())
            self.showMaximized()
        elif is_fullscreen:
            self.logger.info("Applying fullscreen state")
            if target_screen:
                screen_geometry = target_screen.geometry()
                self.move(screen_geometry.x(), screen_geometry.y())
            self.showFullScreen()
        else:
            self.logger.info("Window restored to normal state")
            self.showNormal()

        # Raise the window to the top of the stack
        self.raise_()

    ##### End window properties #####
    #################################

    ###### Window handlers ######
    def on_toggle_tool_signal(self, data: Dict):
        self.toggle_tool(data["tool"], data["active"])

    def on_enable_workflows_toggled(self, message: Dict):
        self._toggle_agent_workflow_feature(message.get("enabled", False))

    def on_retranslate_ui_signal(self):
        self.ui.retranslateUi(self)

    def _toggle_agent_workflow_feature(self, enabled: bool):
        """
        Toggles the visibility of the workflow tab and menu based on the given value.
        """
        # --- Tab handling ---
        tab_widget = self.ui.center_tab_container
        if not self.workflow_tab:
            self.workflow_tab = self.ui.agent_workflow_tab

        if enabled:
            # Only add if not present
            if self.tab_backup != {}:
                tab_widget.addTab(
                    self.tab_backup["tab_widget"],
                    self.tab_backup["tab_text"],
                )
                self.tab_backup = {}
        else:
            # Remove if present
            index = tab_widget.indexOf(self.workflow_tab)
            self.tab_backup = dict(
                tab_text=tab_widget.tabText(index),
                tab_widget=tab_widget.widget(index),
            )
            if index != -1:
                tab_widget.removeTab(index)

        # --- Menu handling ---
        # If menuWorkflow is a QMenu, use menuBar(). If it's an action, use setVisible.
        if hasattr(self.ui, "menuWorkflow"):
            menu = self.ui.menuWorkflow
            # Try both methods for robustness
            try:
                menu.menuAction().setVisible(enabled)
            except Exception:
                try:
                    menu.setVisible(enabled)
                except Exception:
                    pass

    ###### End window handlers ######

    def show_update_message(self):
        if not self.api:
            self.logger.warning(
                "MainWindow: self.api is missing. Cannot display update message."
            )
            return
        self.api.application_status(
            f"New version available: {self.latest_version}"
        )

    def show_update_popup(self):
        self.update_popup = UpdateWindow()

    @staticmethod
    def show_setup_wizard():
        AppInstaller(close_on_cancel=False)

    def showEvent(self, event):
        """Override to update the tray menu text when window is shown."""
        super().showEvent(event)
        # Make sure we update the menu text whenever the window is shown
        if hasattr(self, "toggle_visibility_action"):
            self.toggle_visibility_action.setText("Hide Window")
            # Update the tray menu so it reflects the current state immediately
            self.tray_icon.setContextMenu(self.tray_menu)

        self.logger.debug("showEvent called, initializing window")
        self._initialize_window()
        self._initialize_filter_actions()

        self.initialized = True
        self.logger.debug("Showing window")
        self._set_keyboard_shortcuts()
        
        # Show donation dialog after window is fully displayed (only on first show)
        if not hasattr(self, "_donation_dialog_shown"):
            self._donation_dialog_shown = True
            from PySide6.QtCore import QTimer
            # Show privacy consent first, then donation dialog
            QTimer.singleShot(300, self._show_privacy_consent_dialog)
            QTimer.singleShot(500, self._show_donation_dialog)
    
    def _show_privacy_consent_dialog(self):
        """Show the privacy consent dialog on first launch."""
        from airunner.components.application.gui.dialogs.privacy_consent_dialog import (
            PrivacyConsentDialog,
        )
        PrivacyConsentDialog.show_if_needed(self)

    def _show_donation_dialog(self):
        """Show the donation dialog if appropriate."""
        from airunner.components.application.gui.dialogs.donation_dialog import (
            DonationDialog,
        )
        DonationDialog.show_if_appropriate(self)

    def move_to_second_screen(self):
        screens = QGuiApplication.screens()
        if len(screens) > 1:
            screen = screens[1]
            screen_geometry = screen.availableGeometry()
            self.move(screen_geometry.topLeft())
            self.resize(screen_geometry.size())
            self.setMinimumSize(512, 512)
            max_width = max(screen_geometry.width(), self.minimumWidth())
            max_height = max(screen_geometry.height(), self.minimumHeight())
            self.setMaximumSize(max_width, max_height)

    def on_keyboard_shortcuts_updated(self):
        self._set_keyboard_shortcuts()

    def _set_keyboard_shortcuts(self):
        quit_key = ShortcutKeys.objects.filter_by_first(display_name="Quit")
        if quit_key is not None:
            key_sequence = QKeySequence(quit_key.key | quit_key.modifiers)
            self.ui.actionQuit.setShortcut(key_sequence)
            self.ui.actionQuit.setToolTip(
                f"{quit_key.display_name} ({quit_key.text})"
            )

    def _initialize_filter_actions(self):
        self.ui.menuFilters.clear()
        image_filters = ImageFilter.objects.all()
        try:
            for image_filter in image_filters:
                action = self.ui.menuFilters.addAction(
                    image_filter.display_name
                )
                action.triggered.connect(
                    partial(self.display_filter_window, image_filter)
                )
        except RuntimeError as e:
            if AIRUNNER_ART_ENABLED:
                self.logger.warning(f"Error setting SD status text: {e}")

    @staticmethod
    def display_filter_window(image_filter):
        FilterWindow(image_filter.id)

    def _initialize_window(self):
        # Don't override window geometry if it's already been restored
        if hasattr(self, "_state_restored") and self._state_restored:
            self.logger.debug(
                "Skipping window initialization - state already restored"
            )
            self.setWindowIcon(
                QIcon(
                    os.path.join(
                        self.path_settings.base_path, "images/icon.png"
                    )
                )
            )
            self.set_window_title()
            return

        # self.center()
        screen = QGuiApplication.primaryScreen()  # Use primaryScreen

        if not screen:
            self.logger.warning(
                "Could not get primary screen. Falling back to default size."
            )
            # Fallback to a default size if screen info is unavailable
            default_width, default_height = 1024, 768
            self.resize(default_width, default_height)
            self.setMinimumSize(512, 512)
            # Set maximum size to something reasonable if screen info is missing
            self.setMaximumSize(
                default_width, default_height
            )  # Or a larger sensible max
        else:
            screen_geometry = screen.availableGeometry()
            self.logger.info(
                f"Available screen geometry: "
                f"x={screen_geometry.x()}, y={screen_geometry.y()}, "
                f"width={screen_geometry.width()}, height={screen_geometry.height()}"
            )

            # Set geometry using explicit move and resize
            self.move(screen_geometry.topLeft())
            self.resize(screen_geometry.size())

            self.setMinimumSize(512, 512)
            # Ensure maximum size is at least the minimum size and not smaller than the available geometry
            max_width = max(screen_geometry.width(), self.minimumWidth())
            max_height = max(screen_geometry.height(), self.minimumHeight())
            self.setMaximumSize(max_width, max_height)

        self.setWindowIcon(
            QIcon(
                os.path.join(self.path_settings.base_path, "images/icon.png")
            )
        )
        self.set_window_title()

    def center(self):
        available_geometry = (
            QGuiApplication.primaryScreen().availableGeometry()
        )
        frame_geometry = self.frameGeometry()
        frame_geometry.moveCenter(available_geometry.center())
        self.move(frame_geometry.topLeft())

    def set_window_title(self):
        """
        Overrides base method to set the window title
        :return:
        """
        self.setWindowTitle(self._window_title)

    def handle_unknown(self, message):
        self.logger.error(f"Unknown message code: {message}")

    def clear_all_prompts(self):
        self.prompt = ""
        self.negative_prompt = ""
        if not self.api:
            self.logger.warning(
                "MainWindow: self.api is missing. Cannot clear prompts."
            )
            return
        self.api.clear_prompts()

    def new_batch(self, index, image, data):
        self.generator_tab_widget.new_batch(index, image, data)

    def on_model_status_changed_signal(self, data):
        model = data["model"]
        status = data["status"]
        if self._model_status[model] is status:
            return
        self._model_status[model] = status
        if model is ModelType.SD:
            self.ui.actionToggle_Stable_Diffusion.setDisabled(
                status is ModelStatus.LOADING
            )
            if status is ModelStatus.LOADED:
                self.update_application_settings(sd_enabled=True)
            elif status is ModelStatus.FAILED:
                self.ui.actionToggle_Stable_Diffusion.setChecked(False)
            elif status is ModelStatus.UNLOADED:
                self.update_application_settings(sd_enabled=False)
        elif model is ModelType.LLM:
            self.ui.actionToggle_LLM.setDisabled(status is ModelStatus.LOADING)
            if status is ModelStatus.LOADED:
                self.update_application_settings(llm_enabled=True)
            elif status is ModelStatus.FAILED:
                self.logger.warning("LLM failed to load")
                self.ui.actionToggle_LLM.setChecked(False)
            elif status is ModelStatus.UNLOADED:
                self.update_application_settings(llm_enabled=False)
        elif model is ModelType.TTS:
            self.ui.actionToggle_Text_to_Speech.setDisabled(
                status is ModelStatus.LOADING
            )
            if status is ModelStatus.LOADED:
                self.update_application_settings(tts_enabled=True)
            elif status is ModelStatus.FAILED:
                self.ui.actionToggle_Text_to_Speech.setChecked(False)
            elif status is ModelStatus.UNLOADED:
                self.update_application_settings(tts_enabled=False)
        elif model is ModelType.STT:
            self.ui.actionToggle_Speech_to_Text.setDisabled(
                status is ModelStatus.LOADING
            )
            if status is ModelStatus.LOADED:
                self.update_application_settings(stt_enabled=True)
            elif status is ModelStatus.FAILED:
                self.ui.actionToggle_Speech_to_Text.setChecked(False)
            elif status is ModelStatus.UNLOADED:
                self.update_application_settings(stt_enabled=False)
        self.initialize_widget_elements()
        QApplication.processEvents()

    def _generate_drawingpad_mask(self):
        width = self.application_settings.working_width
        height = self.application_settings.working_height
        img = Image.new("RGB", (width, height), (0, 0, 0))
        base64_image = convert_image_to_binary(img)

        self.update_drawing_pad_settings(mask=base64_image)

    def display_missing_models_error(self, data):
        msg_box = QMessageBox()
        msg_box.setIcon(QMessageBox.Critical)
        msg_box.setWindowTitle(data.get("title", "Error: Missing models"))
        msg_box.setText(data.get("message", "Something went wrong"))
        msg_box.exec()

    def on_status_error_signal(self, data):
        msg_box = QMessageBox()
        msg_box.setIcon(QMessageBox.Critical)
        msg_box.setWindowTitle(data.get("title", "Error"))
        msg_box.setText(data.get("message", "Something went wrong"))
        msg_box.exec()

    def hide_center_tab_header(self):
        """Hide the tab bar for center_tab_container so tabs are not visible."""
        if hasattr(self.ui, "center_tab_container"):
            tab_widget = self.ui.center_tab_container
            tab_bar = tab_widget.tabBar()
            tab_bar.hide()

    def _store_active_tab_index(self, index: int) -> None:
        """Store the active main tab index in QSettings."""
        self.qsettings.beginGroup("window_settings")
        self.qsettings.setValue("active_main_tab_index", index)
        self.qsettings.endGroup()
        self.qsettings.sync()



#!/usr/bin/env python
"""
Test script to check if LoRA adapters are loaded properly.
This helps diagnose if the adapter knowledge is being applied to the model.
"""
import json
from airunner.utils.settings.get_qsettings import get_qsettings
from airunner.components.llm.data.fine_tuned_model import FineTunedModel


def test_adapter_loading():
    """Test adapter configuration and availability."""
    print("=" * 60)
    print("Testing Adapter Loading Configuration")
    print("=" * 60)

    # Check QSettings for enabled adapters
    print("\n1. QSettings Configuration:")
    qs = get_qsettings()
    enabled_json = qs.value("llm_settings/enabled_adapters", "[]")
    print(f"   Enabled adapters JSON: {enabled_json}")

    try:
        enabled_names = json.loads(enabled_json)
        print(f"   Enabled adapter names: {enabled_names}")
    except json.JSONDecodeError as e:
        print(f"   ERROR: Failed to parse JSON: {e}")
        return

    # Check database for adapters
    print("\n2. Database Adapters:")
    try:
        all_adapters = FineTunedModel.objects.all()
        print(f"   Total adapters in database: {len(all_adapters)}")

        for adapter in all_adapters:
            enabled = "✓" if adapter.name in enabled_names else "✗"
            print(f"   {enabled} {adapter.name}")
            print(f"      Path: {adapter.adapter_path}")
            if adapter.adapter_path:
                import os

                exists = os.path.exists(adapter.adapter_path)
                print(f"      Exists: {exists}")
                if exists:
                    # Check if adapter_config.json exists
                    config_path = os.path.join(
                        adapter.adapter_path, "adapter_config.json"
                    )
                    config_exists = os.path.exists(config_path)
                    print(f"      Has adapter_config.json: {config_exists}")
                    if config_exists:
                        with open(config_path, "r") as f:
                            config = json.load(f)
                            print(
                                f"      Adapter type: {config.get('peft_type', 'unknown')}"
                            )
                            print(
                                f"      Task type: {config.get('task_type', 'unknown')}"
                            )
    except Exception as e:
        print(f"   ERROR: {e}")

    # Check which adapters will be loaded
    print("\n3. Adapters That Will Be Loaded:")
    if enabled_names:
        try:
            all_adapters = FineTunedModel.objects.all()
            adapters_to_load = [
                a for a in all_adapters if a.name in enabled_names
            ]
            print(
                f"   Found {len(adapters_to_load)} adapter(s) matching enabled names"
            )
            for adapter in adapters_to_load:
                import os

                if adapter.adapter_path and os.path.exists(
                    adapter.adapter_path
                ):
                    print(f"   ✓ {adapter.name} will be loaded")
                else:
                    print(
                        f"   ✗ {adapter.name} WILL NOT load (path missing or invalid)"
                    )
        except Exception as e:
            print(f"   ERROR: {e}")
    else:
        print("   No adapters enabled")

    print("\n" + "=" * 60)
    print("Test complete.")
    print("\nTo test if adapter knowledge is loaded:")
    print("1. Enable your book-trained adapter in LLM settings")
    print("2. Reload the model (unload and load again)")
    print("3. Check logs for: '✓ Successfully loaded adapter'")
    print("4. Ask the LLM specific questions about your book")
    print("5. Compare responses with/without adapter enabled")
    print("=" * 60)


if __name__ == "__main__":
    test_adapter_loading()



from sqlalchemy import Column, Integer, Boolean, String

from airunner.components.data.models.base import BaseModel
from airunner.enums import ModelService


class RAGSettings(BaseModel):
    __tablename__ = "rag_settings"
    id = Column(Integer, primary_key=True, autoincrement=True)
    model_service = Column(String, default=ModelService.LOCAL.value)
    model_path = Column(String, default="")
    chunk_size = Column(Integer, default=512)
    chunk_overlap = Column(Integer, default=50)



from typing import Dict
from PySide6.QtWidgets import QColorDialog
from PySide6.QtCore import Slot

from airunner.components.application.gui.widgets.base_widget import BaseWidget
from airunner.components.art.gui.widgets.grid_preferences.templates.grid_preferences_ui import (
    Ui_grid_preferences,
)
from airunner.enums import SignalCode


class GridPreferencesWidget(BaseWidget):
    widget_class_ = Ui_grid_preferences

    def __init__(self, *args, **kwargs):
        self.signal_handlers = {
            SignalCode.TOGGLE_GRID: self.on_toggle_grid_signal,
            SignalCode.TOGGLE_GRID_SNAP: self.on_toggle_grid_snap_signal,
        }
        super().__init__(*args, **kwargs)
        self.ui.grid_line_width_spinbox.blockSignals(True)
        self.ui.grid_size_spinbox.blockSignals(True)
        self.ui.show_grid_checkbox.blockSignals(True)
        self.ui.snap_to_grid_checkbox.blockSignals(True)

        self.ui.grid_line_width_spinbox.setValue(self.grid_settings.line_width)
        self.ui.grid_size_spinbox.setValue(self.grid_settings.cell_size)
        self.ui.show_grid_checkbox.setChecked(self.grid_settings.show_grid)
        self.ui.snap_to_grid_checkbox.setChecked(
            self.grid_settings.snap_to_grid
        )

        self.ui.grid_line_width_spinbox.blockSignals(False)
        self.ui.grid_size_spinbox.blockSignals(False)
        self.ui.show_grid_checkbox.blockSignals(False)
        self.ui.snap_to_grid_checkbox.blockSignals(False)

    @Slot(bool)
    def on_snap_to_grid_checkbox_toggled(self, val: bool):
        self.api.art.canvas.toggle_grid_snap(val)

    @Slot(bool)
    def on_show_grid_checkbox_toggled(self, val: bool):
        self.api.art.canvas.toggle_grid(val)

    @Slot()
    def on_grid_line_color_button_clicked(self):
        color = QColorDialog.getColor(parent=self)
        if color.isValid():
            self.grid_settings.line_color = color.name()
            self.update_grid_settings(line_color=color.name())

    @Slot()
    def on_canvas_color_button_clicked(self):
        color = QColorDialog.getColor(parent=self)
        if color.isValid():
            self.grid_settings.canvas_color = color.name()
            self.update_grid_settings(canvas_color=color.name())

    def grid_size_changed(self, val):
        self.grid_settings.cell_size = val
        self.update_grid_settings(cell_size=val)

    def line_width_changed(self, val):
        self.grid_settings.line_width = val
        self.update_grid_settings(line_width=val)

    def on_toggle_grid_signal(self, data: Dict):
        show_grid = data.get("show_grid", False)
        self.ui.show_grid_checkbox.blockSignals(True)
        self.ui.show_grid_checkbox.setChecked(show_grid)
        self.ui.show_grid_checkbox.blockSignals(False)

    def on_toggle_grid_snap_signal(self, data: Dict):
        snap_to_grid = data.get("snap_to_grid", False)
        self.ui.snap_to_grid_checkbox.blockSignals(True)
        self.ui.snap_to_grid_checkbox.setChecked(snap_to_grid)
        self.ui.snap_to_grid_checkbox.blockSignals(False)



"""
Complete model download and quantization pipeline for AI Runner.

This module provides a unified interface for:
1. Downloading models from HuggingFace
2. Quantizing models to 2-bit or 4-bit
3. Cleaning up temporary files
4. Managing model registry
"""

import logging
import os
import shutil
from pathlib import Path
from typing import Optional, Callable
import json

from airunner.components.llm.utils.model_downloader import (
    HuggingFaceDownloader,
)
from airunner.components.llm.utils.model_quantizer import ModelQuantizer
from airunner.settings import AIRUNNER_LOG_LEVEL
from airunner.utils.application import get_logger
from airunner.components.application.gui.windows.main.settings_mixin import (
    SettingsMixin,
)
from airunner.utils.application.mediator_mixin import MediatorMixin

logger = get_logger(__name__, AIRUNNER_LOG_LEVEL)


class ModelPipeline(
    MediatorMixin,
    SettingsMixin,
):
    """
    Unified pipeline for downloading and quantizing LLM models.

    Workflow:
    1. Download model from HuggingFace → temp directory
    2. Quantize to 4-bit (recommended) → final directory
    3. Optionally quantize to 2-bit → final directory
    4. Clean up unquantized files
    5. Update model registry
    """

    # Supported models with their characteristics
    SUPPORTED_MODELS = {
        "ministral-8b": {
            "repo_id": "mistralai/Ministral-8B-Instruct-2410",
            "model_type": "mistral",
            "supports_function_calling": True,
            "context_length": 128000,
            "vram_requirements_gb": {
                "full": 32,
                "4bit": 8,
                "2bit": 4,
            },
            "recommended_bits": 4,
        },
        "qwen2.5-7b": {
            "repo_id": "Qwen/Qwen2.5-7B-Instruct",
            "model_type": "llm",
            "supports_function_calling": True,
            "context_length": 32768,
            "vram_requirements_gb": {
                "full": 28,
                "4bit": 7,
                "2bit": 4,
            },
            "recommended_bits": 4,
        },
        "qwen3-8b": {
            "repo_id": "Qwen/Qwen3-8B-GGUF",
            "gguf_filename": "Qwen3-8B-Q4_K_M.gguf",
            "model_type": "llm",
            "supports_function_calling": True,
            "supports_thinking": True,
            "context_length": 32768,
            "vram_requirements_gb": {
                "full": 32,
                "4bit": 8,
                "2bit": 5,
            },
            "recommended_bits": 4,
        },
    }

    def __init__(self, base_path: Optional[str] = None):
        """
        Initialize pipeline.

        Args:
            base_path: Base directory for models (default: ~/.local/share/airunner/text/models/llm)
        """
        super().__init__()
        if base_path is None:
            base_path = os.path.join(
                os.path.expanduser(self.path_settings.base_path),
                f"text/models/llm",
            )

        self.base_path = Path(base_path)
        self.downloads_dir = (
            self.base_path / "causallm"
        )  # Download directly to causallm
        self.models_dir = self.base_path / "causallm"
        self.registry_file = self.base_path / "registry.json"

        # Create directories
        self.downloads_dir.mkdir(parents=True, exist_ok=True)
        self.models_dir.mkdir(parents=True, exist_ok=True)

        # Initialize components
        self.downloader = HuggingFaceDownloader(
            cache_dir=str(self.downloads_dir)
        )
        self.quantizer = ModelQuantizer()

        # Load or create registry
        self.registry = self._load_registry()

    def download_and_quantize(
        self,
        model_key: str,
        quantize_4bit: bool = True,
        quantize_2bit: bool = False,
        keep_unquantized: bool = False,
        progress_callback: Optional[Callable[[str, float], None]] = None,
    ) -> dict:
        """
        Complete pipeline: download model and create quantized versions.

        Args:
            model_key: Model key from SUPPORTED_MODELS (e.g., "ministral-8b")
            quantize_4bit: Create 4-bit quantized version (recommended)
            quantize_2bit: Create 2-bit quantized version (maximum compression)
            keep_unquantized: Keep full-precision model after quantization
            progress_callback: Callback(stage_description, progress_0_to_1)

        Returns:
            Dict with paths to created model variants
        """
        if model_key not in self.SUPPORTED_MODELS:
            raise ValueError(
                f"Unsupported model: {model_key}. "
                f"Supported: {list(self.SUPPORTED_MODELS.keys())}"
            )

        model_info = self.SUPPORTED_MODELS[model_key]
        repo_id = model_info["repo_id"]
        model_name = repo_id.split("/")[-1]

        logger.info(f"Starting pipeline for {model_key} ({repo_id})")

        results = {}

        # Step 1: Download model
        if progress_callback:
            progress_callback(f"Downloading {model_name}", 0.0)

        download_path = self.downloader.download_model(
            repo_id=repo_id,
            model_type=model_info["model_type"],
            include_patterns=["*.safetensors", "*.json"],
            exclude_patterns=["*.bin", "*.msgpack", "*consolidated*"],
            progress_callback=lambda f, d, t: (
                progress_callback(
                    f"Downloading {f}", 0.0 + (0.4 * d / t) if t > 0 else 0.0
                )
                if progress_callback
                else None
            ),
        )

        results["downloaded"] = str(download_path)

        # Step 2: Quantize to 4-bit
        if quantize_4bit:
            if progress_callback:
                progress_callback(f"Quantizing {model_name} to 4-bit", 0.4)

            output_4bit = self.models_dir / f"{model_name}-4bit-gptq"

            self.quantizer.quantize_model(
                model_path=str(download_path),
                output_path=str(output_4bit),
                bits=4,
                progress_callback=lambda stage, pct: (
                    progress_callback(f"4-bit: {stage}", 0.4 + (0.3 * pct))
                    if progress_callback
                    else None
                ),
            )

            results["4bit"] = str(output_4bit)

            # Update registry
            self._add_to_registry(
                model_key=f"{model_key}-4bit",
                path=str(output_4bit),
                model_info=model_info,
                quantization="4bit",
            )

        # Step 3: Quantize to 2-bit (optional)
        if quantize_2bit:
            if progress_callback:
                progress_callback(f"Quantizing {model_name} to 2-bit", 0.7)

            output_2bit = self.models_dir / f"{model_name}-2bit-gptq"

            self.quantizer.quantize_model(
                model_path=str(download_path),
                output_path=str(output_2bit),
                bits=2,
                progress_callback=lambda stage, pct: (
                    progress_callback(f"2-bit: {stage}", 0.7 + (0.2 * pct))
                    if progress_callback
                    else None
                ),
            )

            results["2bit"] = str(output_2bit)

            # Update registry
            self._add_to_registry(
                model_key=f"{model_key}-2bit",
                path=str(output_2bit),
                model_info=model_info,
                quantization="2bit",
            )

        # Step 4: Clean up unquantized model (unless requested to keep)
        if not keep_unquantized:
            if progress_callback:
                progress_callback("Cleaning up temporary files", 0.9)

            logger.info(f"Removing unquantized model: {download_path}")
            shutil.rmtree(download_path)
        else:
            results["unquantized"] = str(download_path)

        if progress_callback:
            progress_callback("Complete", 1.0)

        logger.info(f"Pipeline complete for {model_key}")
        logger.info(f"Results: {results}")

        return results

    def download_from_url(
        self,
        url: str,
        quantize_4bit: bool = True,
        quantize_2bit: bool = False,
        progress_callback: Optional[Callable[[str, float], None]] = None,
    ) -> dict:
        """
        Download and quantize a model from a HuggingFace URL.

        Args:
            url: HuggingFace model URL (e.g., "https://huggingface.co/mistralai/Ministral-8B-Instruct-2410")
            quantize_4bit: Create 4-bit version
            quantize_2bit: Create 2-bit version
            progress_callback: Callback(stage_description, progress_0_to_1)

        Returns:
            Dict with paths to created model variants
        """
        # Extract repo_id from URL
        # URL format: https://huggingface.co/{owner}/{model_name}
        if "//" in url:
            url = url.split("//")[1]  # Remove protocol

        parts = url.split("/")
        if "huggingface.co" in parts:
            idx = parts.index("huggingface.co")
            repo_id = "/".join(parts[idx + 1 : idx + 3])
        else:
            raise ValueError(f"Invalid HuggingFace URL: {url}")

        logger.info(f"Extracted repo_id: {repo_id}")

        # Check if it's a known model
        for key, info in self.SUPPORTED_MODELS.items():
            if info["repo_id"] == repo_id:
                logger.info(f"Recognized model: {key}")
                return self.download_and_quantize(
                    model_key=key,
                    quantize_4bit=quantize_4bit,
                    quantize_2bit=quantize_2bit,
                    progress_callback=progress_callback,
                )

        # Unknown model - try to download anyway (assume "llm" type)
        logger.warning(f"Unknown model: {repo_id}. Attempting download...")

        model_name = repo_id.split("/")[-1]

        # Download
        download_path = self.downloader.download_model(
            repo_id=repo_id,
            model_type="llm",
            include_patterns=["*.safetensors", "*.json"],
            exclude_patterns=["*.bin", "*.msgpack", "*consolidated*"],
            progress_callback=lambda f, d, t: (
                progress_callback(
                    f"Downloading {f}", 0.0 + (0.5 * d / t) if t > 0 else 0.0
                )
                if progress_callback
                else None
            ),
        )

        results = {"downloaded": str(download_path)}

        # Quantize if requested
        if quantize_4bit:
            output_4bit = self.models_dir / f"{model_name}-4bit-gptq"
            self.quantizer.quantize_model(
                model_path=str(download_path),
                output_path=str(output_4bit),
                bits=4,
                progress_callback=lambda stage, pct: (
                    progress_callback(f"4-bit: {stage}", 0.5 + (0.5 * pct))
                    if progress_callback
                    else None
                ),
            )
            results["4bit"] = str(output_4bit)

        return results

    def list_available_models(self) -> dict:
        """
        List all supported models and their requirements.

        Returns:
            Dict of model info
        """
        return self.SUPPORTED_MODELS.copy()

    def list_downloaded_models(self) -> list:
        """
        List models that have been downloaded.

        Returns:
            List of model info dicts from registry
        """
        return list(self.registry.values())

    def _load_registry(self) -> dict:
        """Load model registry from disk."""
        if self.registry_file.exists():
            try:
                with open(self.registry_file) as f:
                    return json.load(f)
            except Exception as e:
                logger.error(f"Failed to load registry: {e}")
                return {}
        return {}

    def _save_registry(self):
        """Save model registry to disk."""
        try:
            with open(self.registry_file, "w") as f:
                json.dump(self.registry, f, indent=2)
        except Exception as e:
            logger.error(f"Failed to save registry: {e}")

    def _add_to_registry(
        self,
        model_key: str,
        path: str,
        model_info: dict,
        quantization: str,
    ):
        """Add a model to the registry."""
        self.registry[model_key] = {
            "path": path,
            "repo_id": model_info["repo_id"],
            "quantization": quantization,
            "supports_function_calling": model_info[
                "supports_function_calling"
            ],
            "context_length": model_info["context_length"],
            "vram_gb": model_info["vram_requirements_gb"][quantization],
        }
        self._save_registry()


# Example usage
if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    pipeline = ModelPipeline()

    # List available models
    print("Available models:")
    for key, info in pipeline.list_available_models().items():
        print(f"  {key}: {info['repo_id']}")
        print(f"    Function calling: {info['supports_function_calling']}")
        print(f"    VRAM (4-bit): {info['vram_requirements_gb']['4bit']} GB")

    # Download and quantize Ministral-8B
    def progress(stage, pct):
        print(f"[{pct * 100:5.1f}%] {stage}")

    results = pipeline.download_and_quantize(
        model_key="ministral-8b",
        quantize_4bit=True,
        quantize_2bit=False,
        progress_callback=progress,
    )

    print("\nResults:", results)



"""Conditional Branch Node for routing in LangGraph workflows.

This node creates conditional routing based on state.
"""

from typing import Dict
from airunner.vendor.nodegraphqt.constants import NodePropWidgetEnum
from airunner.components.nodegraph.gui.widgets.nodes.langgraph.base_langgraph_node import (
    BaseLangGraphNode,
)


class ConditionalBranchNode(BaseLangGraphNode):
    """Conditional routing node.

    This node evaluates a condition and routes to different
    nodes based on the result.
    """

    NODE_NAME = "Conditional Branch"
    is_conditional = True

    # Conditional nodes have multiple output exec ports
    has_exec_out_port = False  # We'll add custom ones

    _input_ports = [
        dict(name="condition_value", display_name="Condition Value"),
    ]

    _output_ports = [
        dict(name="true_output", display_name="True"),
        dict(name="false_output", display_name="False"),
    ]

    _properties = [
        dict(
            name="condition_code",
            value='return state.get("next_action") == "continue"',
            widget_type=NodePropWidgetEnum.QTEXT_EDIT,
            tab="condition",
        ),
        dict(
            name="true_route",
            value="continue",
            widget_type=NodePropWidgetEnum.QLINE_EDIT,
            tab="routing",
        ),
        dict(
            name="false_route",
            value="END",
            widget_type=NodePropWidgetEnum.QLINE_EDIT,
            tab="routing",
        ),
    ]

    def __init__(self):
        """Initialize conditional branch node."""
        super().__init__()
        # Add execution output ports for true/false branches
        self.add_output(
            "exec_out_true",
            multi_output=False,
            display_name="True",
            color=(0, 255, 0),
            painter_func=self._draw_exec_port,
        )
        self.add_output(
            "exec_out_false",
            multi_output=False,
            display_name="False",
            color=(255, 0, 0),
            painter_func=self._draw_exec_port,
        )

    def get_node_type(self) -> str:
        """Get node type identifier."""
        return "conditional"

    def get_description(self) -> str:
        """Get node description."""
        return "Conditional routing"

    def to_langgraph_code(self) -> str:
        """Generate Python code for conditional routing.

        Returns:
            Python code string
        """
        condition_code = self.get_property("condition_code")
        true_route = self.get_property("true_route")
        false_route = self.get_property("false_route")

        func_name = self._sanitize_name(self.name())

        code = f'''def {func_name}_condition(state: AgentState) -> str:
    """Conditional routing for {self.name()}"""
    try:
        # Evaluate condition
        {condition_code}
        result = {condition_code}
        
        if result:
            return "{true_route}"
        else:
            return "{false_route}"
    
    except Exception as e:
        logger.error(f"Condition evaluation error: {{e}}")
        return "{false_route}"'''

        return code

    def get_condition_mapping(self) -> Dict[str, str]:
        """Get the routing mapping for this conditional.

        Returns:
            Dict mapping condition results to node names
        """
        true_route = self.get_property("true_route")
        false_route = self.get_property("false_route")

        # Find connected nodes
        true_port = None
        false_port = None

        for port in self.output_ports():
            if "true" in port.name().lower():
                true_port = port
            elif "false" in port.name().lower():
                false_port = port

        mapping = {}

        if true_port and true_port.connected_ports():
            target_node = true_port.connected_ports()[0].node()
            mapping[true_route] = target_node.name()

        if false_port and false_port.connected_ports():
            target_node = false_port.connected_ports()[0].node()
            mapping[false_route] = target_node.name()

        return mapping

    @staticmethod
    def _sanitize_name(name: str) -> str:
        """Sanitize node name to valid Python identifier."""
        sanitized = "".join(
            c if c.isalnum() or c == "_" else "_" for c in name
        )
        if sanitized and sanitized[0].isdigit():
            sanitized = f"node_{sanitized}"
        return sanitized or "condition_node"



"""
Mixins for BaseDiffusersModelManager.

These mixins break down the large BaseDiffusersModelManager class into
focused, single-responsibility components for improved maintainability.
"""

from airunner.components.art.managers.stablediffusion.mixins.sd_properties_mixin import (
    SDPropertiesMixin,
)
from airunner.components.art.managers.stablediffusion.mixins.sd_pipeline_management_mixin import (
    SDPipelineManagementMixin,
)
from airunner.components.art.managers.stablediffusion.mixins.sd_model_loading_mixin import (
    SDModelLoadingMixin,
)
from airunner.components.art.managers.stablediffusion.mixins.sd_model_unloading_mixin import (
    SDModelUnloadingMixin,
)
from airunner.components.art.managers.stablediffusion.mixins.sd_memory_management_mixin import (
    SDMemoryManagementMixin,
)
from airunner.components.art.managers.stablediffusion.mixins.sd_generation_preparation_mixin import (
    SDGenerationPreparationMixin,
)
from airunner.components.art.managers.stablediffusion.mixins.sd_image_generation_mixin import (
    SDImageGenerationMixin,
)

__all__ = [
    "SDPropertiesMixin",
    "SDPipelineManagementMixin",
    "SDModelLoadingMixin",
    "SDModelUnloadingMixin",
    "SDMemoryManagementMixin",
    "SDGenerationPreparationMixin",
    "SDImageGenerationMixin",
]



from abc import ABC, abstractmethod
from typing import Tuple, Type, Optional, Dict, Callable
import threading

import pygame
from pygame.locals import *

from PySide6.QtWidgets import QMainWindow, QSizePolicy
from PySide6.QtWidgets import QWidget
from PySide6.QtCore import QTimer, Qt, QSize
from PySide6.QtGui import (
    QResizeEvent,
    QPainter,
    QImage,
    QKeyEvent,
    QMouseEvent,
)
from PySide6.QtWidgets import QVBoxLayout

from airunner.enums import SignalCode
from airunner.components.stt.workers.audio_capture_worker import AudioCaptureWorker
from airunner.components.stt.workers.audio_processor_worker import AudioProcessorWorker
from airunner.components.llm.workers.llm_generate_worker import LLMGenerateWorker
from airunner.components.art.workers.mask_generator_worker import MaskGeneratorWorker
from airunner.components.art.workers.sd_worker import SDWorker
from airunner.components.tts.workers.tts_generator_worker import TTSGeneratorWorker
from airunner.components.tts.workers.tts_vocalizer_worker import TTSVocalizerWorker
from airunner.utils.application import create_worker
from airunner.components.application.gui.windows.main.ai_model_mixin import AIModelMixin
from airunner.components.application.gui.windows.main.pipeline_mixin import PipelineMixin
from airunner.components.application.gui.windows.main.settings_mixin import SettingsMixin
from airunner.gui.styles.styles_mixin import StylesMixin
from airunner.utils.application.mediator_mixin import MediatorMixin
from airunner.components.llm.managers.llm_response import LLMResponse
from airunner.components.application.api.api import API
from airunner.components.llm.managers.agent import LocalAgent
from airunner.components.art.managers.stablediffusion.image_response import (
    ImageResponse,
)
from airunner.settings import (
    AIRUNNER_STT_ON,
    AIRUNNER_TTS_ON,
    AIRUNNER_LLM_ON,
    AIRUNNER_SD_ON,
)
from airunner.enums import EngineResponseCode


# Map of PySide6 key codes to Pygame key codes
# This is incomplete and should be expanded as needed
KEY_MAP = {
    Qt.Key.Key_Escape: K_ESCAPE,
    Qt.Key.Key_Return: K_RETURN,
    Qt.Key.Key_Enter: K_RETURN,
    Qt.Key.Key_Space: K_SPACE,
    Qt.Key.Key_Up: K_UP,
    Qt.Key.Key_Down: K_DOWN,
    Qt.Key.Key_Left: K_LEFT,
    Qt.Key.Key_Right: K_RIGHT,
    Qt.Key.Key_A: K_a,
    Qt.Key.Key_B: K_b,
    Qt.Key.Key_C: K_c,
    Qt.Key.Key_D: K_d,
    Qt.Key.Key_E: K_e,
    Qt.Key.Key_F: K_f,
    Qt.Key.Key_G: K_g,
    Qt.Key.Key_H: K_h,
    Qt.Key.Key_I: K_i,
    Qt.Key.Key_J: K_j,
    Qt.Key.Key_K: K_k,
    Qt.Key.Key_L: K_l,
    Qt.Key.Key_M: K_m,
    Qt.Key.Key_N: K_n,
    Qt.Key.Key_O: K_o,
    Qt.Key.Key_P: K_p,
    Qt.Key.Key_Q: K_q,
    Qt.Key.Key_R: K_r,
    Qt.Key.Key_S: K_s,
    Qt.Key.Key_T: K_t,
    Qt.Key.Key_U: K_u,
    Qt.Key.Key_V: K_v,
    Qt.Key.Key_W: K_w,
    Qt.Key.Key_X: K_x,
    Qt.Key.Key_Y: K_y,
    Qt.Key.Key_Z: K_z,
    Qt.Key.Key_0: K_0,
    Qt.Key.Key_1: K_1,
    Qt.Key.Key_2: K_2,
    Qt.Key.Key_3: K_3,
    Qt.Key.Key_4: K_4,
    Qt.Key.Key_5: K_5,
    Qt.Key.Key_6: K_6,
    Qt.Key.Key_7: K_7,
    Qt.Key.Key_8: K_8,
    Qt.Key.Key_9: K_9,
}


class PygameManager(ABC):
    """
    A class to manage the Pygame window and its initialization.
    This class is responsible for setting up the Pygame environment,
    creating the window, and handling events.
    It is designed to be used with the AI Runner framework.
    This class is a wrapper around the Pygame library to provide
    a consistent interface for creating and managing a Pygame window.
    """

    def __init__(
        self,
        api: API,
        game_title: str = "Pygame Window",
        width: int = 800,
        height: int = 600,
        screen_color: Tuple = (0, 0, 0),
        fps: int = 60,
    ):
        self.api = api
        self.game_title: str = game_title
        self.width: int = width
        self.height: int = height
        self.screen: pygame.Surface = None
        self.screen_color: Tuple = screen_color
        self.running: bool = False
        self.fps: int = fps
        self.clock = pygame.time.Clock()
        self._initialize()
        self._start()

        for signal, handler in [
            (
                SignalCode.LLM_TEXT_STREAMED_SIGNAL,
                self._handle_llm_response_signal,
            ),
            (
                SignalCode.ENGINE_RESPONSE_WORKER_RESPONSE_SIGNAL,
                self._handle_image_response_signal,
            ),
        ]:
            self.api.register(signal, handler)

        self.api.logger.info("PygameManager initialized")

    def _initialize(self):
        self.api.logger.info("Initializing Pygame")
        self._initialize_pygame()
        self._initialize_screen()
        self._initialize_display()

    def _handle_llm_response_signal(self, data: Dict):
        response = data.get("response")
        thread = threading.Thread(
            target=self._handle_llm_response, args=(response,)
        )
        thread.start()

    def _handle_image_response_signal(self, data: Dict):
        code = data["code"]
        callback = data.get("callback", None)

        if code in (EngineResponseCode.INSUFFICIENT_GPU_MEMORY,):
            self.api.emit_signal(
                SignalCode.APPLICATION_STATUS_ERROR_SIGNAL,
                "Insufficient GPU memory.",
            )
        elif code is EngineResponseCode.IMAGE_GENERATED:
            thread = threading.Thread(
                target=self._handle_image_response,
                args=(data.get("message", None),),
            )
            thread.start()
        else:
            self.api.logger.error(f"Unhandled response code: {code}")

        self.api.emit_signal(
            SignalCode.APPLICATION_STOP_SD_PROGRESS_BAR_SIGNAL
        )

        if callback:
            callback(data)

    @abstractmethod
    def _handle_llm_response(self, response: LLMResponse):
        """
        Handle the LLM response.
        This method should be overridden by subclasses to provide
        specific functionality for handling LLM responses.
        """

    @abstractmethod
    def _handle_image_response(self, response: Optional[ImageResponse]):
        """
        Handle the image response.
        This method should be overridden by subclasses to provide
        specific functionality for handling image responses.
        """

    @abstractmethod
    def _start(self):
        """
        Start the Pygame loop.
        This method should be overridden by subclasses to provide
        specific functionality for the game loop.
        """

    def process_events(self):
        """
        Process Pygame events in a single step.
        This method should process events but not contain a loop.
        It will be called by the Qt timer.
        """

    def update(self):
        """
        Update game state in a single step.
        This method should be overridden by subclasses to provide
        specific functionality for updating the game state.
        """

    def render(self):
        """
        Render the game state to the screen.
        This method should be overridden by subclasses to provide
        specific functionality for rendering the game state.
        """

    def handle_pygame_event(self, event):
        """
        Handle a single pygame event.
        This method should be overridden by subclasses to provide
        specific functionality for handling pygame events.
        """

    @abstractmethod
    def quit(self):
        """
        Quit the Pygame loop.
        This method should be overridden by subclasses to provide
        specific functionality for quitting the game loop.
        """

    def _initialize_pygame(self):
        """
        Initialize the pygame library.
        """
        if not pygame.get_init():
            pygame.init()

    def _initialize_screen(self):
        """
        Initialize the screen for the pygame window.
        """
        # Use pygame.Surface instead of pygame.display.set_mode
        # This creates an in-memory surface without a window
        self.screen = pygame.Surface((self.width, self.height))

    def _initialize_display(self):
        """
        Initialize the display for the pygame window.
        """
        # We don't set the caption since we're running in a PySide6 window

    def resize(self, width, height):
        """
        Resize the pygame surface.
        """
        self.width = width
        self.height = height
        self.screen = pygame.Surface((width, height))


class PygameWidget(QWidget):
    def __init__(self, pygame_manager, parent=None):
        super().__init__(parent)
        self.pygame_manager = pygame_manager
        self.setAttribute(Qt.WidgetAttribute.WA_OpaquePaintEvent)
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        self.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
        )
        self.setMinimumSize(pygame_manager.width, pygame_manager.height)

        # Timer to update the Pygame display
        self.timer = QTimer(self)
        self.timer.timeout.connect(self.update_pygame)
        self.timer.start(
            int(1000 / self.pygame_manager.fps)
        )  # Use the FPS from pygame_manager

    def update_pygame(self):
        """
        Update the Pygame display and repaint the widget.
        """
        # Process Pygame events in the Qt main thread
        self.pygame_manager.process_events()

        # Update game state
        self.pygame_manager.update()

        # Render the Pygame surface
        self.pygame_manager.render()

        # Maintain the target framerate
        self.pygame_manager.clock.tick(self.pygame_manager.fps)

        # Trigger a repaint
        self.update()

    def paintEvent(self, event):
        """
        Override the paint event to draw the Pygame surface onto the widget.
        """
        if not self.pygame_manager.screen:
            return

        # Create a QPainter for drawing onto this widget
        painter = QPainter(self)

        # Convert Pygame surface to QImage
        surface_data = pygame.image.tostring(self.pygame_manager.screen, "RGB")
        w = self.pygame_manager.screen.get_width()
        h = self.pygame_manager.screen.get_height()

        # Create a QImage from the raw data
        qimage = QImage(surface_data, w, h, w * 3, QImage.Format.Format_RGB888)

        # Draw the QImage onto the widget, scaled to fit if needed
        painter.drawImage(self.rect(), qimage)
        painter.end()

    def resizeEvent(self, event: QResizeEvent):
        """
        Handle resizing of the widget and update the Pygame surface size.
        """
        size = event.size()
        # Resize the pygame surface
        self.pygame_manager.resize(size.width(), size.height())
        super().resizeEvent(event)

    def keyPressEvent(self, event: QKeyEvent):
        """
        Convert Qt key events to Pygame key events
        """
        key = event.key()
        if key in KEY_MAP:
            # Create a pygame KEYDOWN event
            pygame_event = pygame.event.Event(
                KEYDOWN,
                key=KEY_MAP[key],
                mod=self._convert_modifiers(event.modifiers()),
            )
            # Post the event to pygame's event queue
            pygame.event.post(pygame_event)
            # Also let the manager handle it directly
            self.pygame_manager.handle_pygame_event(pygame_event)
        super().keyPressEvent(event)

    def keyReleaseEvent(self, event: QKeyEvent):
        """
        Convert Qt key release events to Pygame key events
        """
        key = event.key()
        if key in KEY_MAP:
            # Create a pygame KEYUP event
            pygame_event = pygame.event.Event(
                KEYUP,
                key=KEY_MAP[key],
                mod=self._convert_modifiers(event.modifiers()),
            )
            # Post the event to pygame's event queue
            pygame.event.post(pygame_event)
            # Also let the manager handle it directly
            self.pygame_manager.handle_pygame_event(pygame_event)
        super().keyReleaseEvent(event)

    def mousePressEvent(self, event: QMouseEvent):
        """
        Convert Qt mouse events to Pygame mouse events
        """
        pos = (event.position().x(), event.position().y())
        button = self._convert_mouse_button(event.button())

        # Create a pygame MOUSEBUTTONDOWN event
        pygame_event = pygame.event.Event(
            MOUSEBUTTONDOWN, pos=pos, button=button
        )
        # Post the event to pygame's event queue
        pygame.event.post(pygame_event)
        # Also let the manager handle it directly
        self.pygame_manager.handle_pygame_event(pygame_event)
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event: QMouseEvent):
        """
        Convert Qt mouse release events to Pygame mouse events
        """
        pos = (event.position().x(), event.position().y())
        button = self._convert_mouse_button(event.button())

        # Create a pygame MOUSEBUTTONUP event
        pygame_event = pygame.event.Event(
            MOUSEBUTTONUP, pos=pos, button=button
        )
        # Post the event to pygame's event queue
        pygame.event.post(pygame_event)
        # Also let the manager handle it directly
        self.pygame_manager.handle_pygame_event(pygame_event)
        super().mouseReleaseEvent(event)

    def mouseMoveEvent(self, event: QMouseEvent):
        """
        Convert Qt mouse move events to Pygame mouse events
        """
        pos = (event.position().x(), event.position().y())
        buttons = self._convert_mouse_buttons(event.buttons())

        # Create a pygame MOUSEMOTION event
        pygame_event = pygame.event.Event(
            MOUSEMOTION,
            pos=pos,
            rel=(0, 0),  # relative movement, we're not tracking this
            buttons=buttons,
        )
        # Post the event to pygame's event queue
        pygame.event.post(pygame_event)
        # Also let the manager handle it directly
        self.pygame_manager.handle_pygame_event(pygame_event)
        super().mouseMoveEvent(event)

    def _convert_modifiers(self, qt_modifiers):
        """
        Convert Qt keyboard modifiers to Pygame modifiers
        """
        pygame_modifiers = 0
        if qt_modifiers & Qt.KeyboardModifier.ShiftModifier:
            pygame_modifiers |= pygame.KMOD_SHIFT
        if qt_modifiers & Qt.KeyboardModifier.ControlModifier:
            pygame_modifiers |= pygame.KMOD_CTRL
        if qt_modifiers & Qt.KeyboardModifier.AltModifier:
            pygame_modifiers |= pygame.KMOD_ALT
        if qt_modifiers & Qt.KeyboardModifier.MetaModifier:
            pygame_modifiers |= pygame.KMOD_META
        return pygame_modifiers

    def _convert_mouse_button(self, qt_button):
        """
        Convert Qt mouse button to Pygame mouse button
        """
        if qt_button == Qt.MouseButton.LeftButton:
            return 1
        elif qt_button == Qt.MouseButton.MiddleButton:
            return 2
        elif qt_button == Qt.MouseButton.RightButton:
            return 3
        return 0

    def _convert_mouse_buttons(self, qt_buttons):
        """
        Convert Qt mouse buttons to Pygame mouse buttons tuple
        """
        left = bool(qt_buttons & Qt.MouseButton.LeftButton)
        middle = bool(qt_buttons & Qt.MouseButton.MiddleButton)
        right = bool(qt_buttons & Qt.MouseButton.RightButton)
        return (left, middle, right)

    def closeEvent(self, event):
        """
        Handle the widget close event and quit Pygame.
        """
        self.pygame_manager.quit()
        if pygame.get_init():
            pygame.quit()
        super().closeEvent(event)

    def sizeHint(self) -> QSize:
        """
        Return the recommended size for the widget.
        """
        return QSize(self.pygame_manager.width, self.pygame_manager.height)


class PygameWindow(
    MediatorMixin,
    SettingsMixin,
    StylesMixin,
    PipelineMixin,
    AIModelMixin,
    QMainWindow,
):
    """
    A class to create a Pygame window using PySide6.
    This class is a subclass of QMainWindow and is designed to be used
    with the AI Runner framework. It initializes a Pygame window and
    manages the Pygame event loop. The class also provides methods for
    sending requests and handling signals.
    """

    def __init__(
        self,
        app,
        game_class: Type[PygameManager],
        width: int = 800,
        height: int = 600,
        local_agent_class: Optional[Type[LocalAgent]] = None,
        fps: int = 60,
        *args,
        **kwargs,
    ):
        """
        Initialize the Pygame window.
        :param app: The main application instance.
        :param game_class: The class that manages the Pygame window.
        :param width: The width of the Pygame window.
        :param height: The height of the Pygame window.
        :param fps: Frames per second for the game loop.
        :param args: Additional arguments.
        :param kwargs: Additional keyword arguments.
        """
        self.app = app
        self.pygame_manager = game_class(
            api=app,
            width=width,
            height=height,
            fps=fps,
            **kwargs.get("pygame_params", {}),
        )
        if AIRUNNER_SD_ON:
            self._mask_generator_worker = create_worker(MaskGeneratorWorker)
            self._sd_worker = create_worker(SDWorker)

        if AIRUNNER_STT_ON:
            self._stt_audio_capture_worker = create_worker(AudioCaptureWorker)
            self._stt_audio_processor_worker = create_worker(
                AudioProcessorWorker
            )

        if AIRUNNER_TTS_ON:
            self._tts_generator_worker = create_worker(TTSGeneratorWorker)
            self._tts_vocalizer_worker = create_worker(TTSVocalizerWorker)

        if AIRUNNER_LLM_ON:
            self._llm_generate_worker = create_worker(
                LLMGenerateWorker, local_agent_class=local_agent_class
            )

        super().__init__(*args, **kwargs)
        self.setWindowTitle(f"AI Runner - {self.pygame_manager.game_title}")
        self.resize(width, height)

        # Create a central widget and layout
        central_widget = QWidget(self)
        layout = QVBoxLayout(central_widget)
        layout.setContentsMargins(
            0, 0, 0, 0
        )  # Remove margins for better display

        # Create and add the PygameWidget
        self.pygame_widget = PygameWidget(self.pygame_manager, self)
        layout.addWidget(self.pygame_widget)

        self.setCentralWidget(central_widget)

        # Mark the game as running
        self.pygame_manager.running = True

        # Make sure the window is visible and has focus
        self.show()
        self.activateWindow()
        self.pygame_widget.setFocus()

        # Start the process to keep the game running
        QTimer.singleShot(0, self.pygame_manager.run)

    def closeEvent(self, event):
        """
        Handle the window close event.
        """
        self.pygame_manager.quit()
        if pygame.get_init():
            pygame.quit()
        super().closeEvent(event)


# Adapter for existing Pygame projects
class PygameAdapter(PygameManager):
    """
    An adapter class that makes it easier to integrate existing Pygame projects.

    This class provides a simpler interface for integrating existing Pygame
    projects with the AI Runner framework. It handles the basic setup and
    provides hooks for the existing game loop.

    Example usage:

    ```python
    # Original Pygame code
    def main():
        pygame.init()
        screen = pygame.display.set_mode((800, 600))
        clock = pygame.time.Clock()
        running = True

        while running:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    running = False
                # Handle other events...

            # Update game state...

            # Render...
            screen.fill((0, 0, 0))
            # Draw stuff...
            pygame.display.flip()

            clock.tick(60)

        pygame.quit()

    # Using the adapter
    class MyGame(PygameAdapter):
        def handle_pygame_event(self, event):
            # Handle events same as in original code
            pass

        def update(self):
            # Update game state same as in original code
            pass

        def render(self):
            # Render same as in original code
            self.screen.fill((0, 0, 0))
            # Draw stuff...
    ```
    """

    def __init__(
        self,
        api: API,
        game_title: str = "Pygame Game",
        width: int = 800,
        height: int = 600,
        screen_color: Tuple = (0, 0, 0),
        fps: int = 60,
        event_handler: Optional[Callable[[pygame.event.Event], None]] = None,
        update_handler: Optional[Callable[[], None]] = None,
        render_handler: Optional[Callable[[pygame.Surface], None]] = None,
    ):
        self.event_handler = event_handler
        self.update_handler = update_handler
        self.render_handler = render_handler
        super().__init__(
            api=api,
            game_title=game_title,
            width=width,
            height=height,
            screen_color=screen_color,
            fps=fps,
        )

    def _handle_llm_response(self, response: LLMResponse):
        # Default implementation - can be overridden
        print(f"LLM: {response.message}")

    def _handle_image_response(self, response: Optional[ImageResponse]):
        # Default implementation - can be overridden
        if response and response.images:
            print(f"Received {len(response.images)} images")

    def _start(self):
        # Default initialization - can be overridden
        self.screen.fill(self.screen_color)

    def process_events(self):
        # Process all pending Pygame events
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.running = False
            self.handle_pygame_event(event)

    def handle_pygame_event(self, event):
        # Use the provided event handler if available
        if self.event_handler:
            self.event_handler(event)

    def update(self):
        # Use the provided update handler if available
        if self.update_handler:
            self.update_handler()

    def render(self):
        # Use the provided render handler if available
        if self.render_handler:
            self.render_handler(self.screen)

    def quit(self):
        self.running = False

    def run(self):
        pass



import re

from airunner.vendor.melo.text import symbols
from airunner.vendor.melo.text.es_phonemizer import cleaner as es_cleaner
from airunner.vendor.melo.text.es_phonemizer import es_to_ipa
from airunner.vendor.melo.text.language_base import LanguageBase


class Spanish(LanguageBase):
    model_path = "dccuchile/bert-base-spanish-wwm-uncased"
    model_path_bert = "dccuchile/bert-base-spanish-wwm-uncased"

    def text_normalize(self, text):
        text = es_cleaner.spanish_cleaners(text)
        return text

    def post_replace_ph(self, ph):
        rep_map = {
            "：": ",",
            "；": ",",
            "，": ",",
            "。": ".",
            "！": "!",
            "？": "?",
            "\n": ".",
            "·": ",",
            "、": ",",
            "...": "…",
        }
        if ph in rep_map.keys():
            ph = rep_map[ph]
        if ph in symbols:
            return ph
        if ph not in symbols:
            ph = "UNK"
        return ph

    def refine_ph(self, phn):
        tone = 0
        if re.search(r"\d$", phn):
            tone = int(phn[-1]) + 1
            phn = phn[:-1]
        return phn.lower(), tone

    def refine_syllables(self, syllables):
        tones = []
        phonemes = []
        for phn_list in syllables:
            for i in range(len(phn_list)):
                phn = phn_list[i]
                phn, tone = self.refine_ph(phn)
                phonemes.append(phn)
                tones.append(tone)
        return phonemes, tones

    def call(self, text, pad_start_end=True, tokenized=None):
        if tokenized is None:
            tokenized = self.tokenizer.tokenize(text)
        # import pdb; pdb.set_trace()
        ph_groups = []
        for t in tokenized:
            if not t.startswith("#"):
                ph_groups.append([t])
            else:
                ph_groups[-1].append(t.replace("#", ""))

        phones = []
        tones = []
        word2ph = []
        # print(ph_groups)
        for group in ph_groups:
            w = "".join(group)
            phone_len = 0
            word_len = len(group)
            if w == "[UNK]":
                phone_list = ["UNK"]
            else:
                phone_list = list(
                    filter(lambda p: p != " ", es_to_ipa.es2ipa(w))
                )

            for ph in phone_list:
                phones.append(ph)
                tones.append(0)
                phone_len += 1
            aaa = self.distribute_phone(phone_len, word_len)
            word2ph += aaa
            # print(phone_list, aaa)
            # print('=' * 10)

        if pad_start_end:
            phones = ["_"] + phones + ["_"]
            tones = [0] + tones + [0]
            word2ph = [1] + word2ph + [1]
        return phones, tones, word2ph



# Signal handlers for NodeGraphWidget

from typing import Dict


def _on_nodegraph_zoom_changed(self, data: Dict):
    """Signal handler for NODEGRAPH_ZOOM signal."""
    zoom = data.get("zoom_level", 0)

    # Get the center from the current application settings
    # Don't try to access viewer directly from this handler
    settings = self.application_settings
    try:
        center_x = int(getattr(settings, "nodegraph_center_x", 0) or 0)
        center_y = int(getattr(settings, "nodegraph_center_y", 0) or 0)
    except (TypeError, ValueError):
        center_x = 0
        center_y = 0

    # Update the settings
    from airunner.components.settings.data.application_settings import ApplicationSettings

    ApplicationSettings.objects.update(
        self.application_settings.id,
        nodegraph_zoom=zoom,
        nodegraph_center_x=center_x,
        nodegraph_center_y=center_y,
    )


def _on_nodegraph_pan_changed(self, data: Dict):
    """Signal handler for NODEGRAPH_PAN signal."""
    try:
        center_x = int(data.get("center_x", 0) or 0)
        center_y = int(data.get("center_y", 0) or 0)
    except (TypeError, ValueError):
        center_x = 0
        center_y = 0

    # Get current zoom from the current application settings
    settings = self.application_settings
    zoom = getattr(settings, "nodegraph_zoom", 0)

    # Update the settings
    from airunner.components.settings.data.application_settings import ApplicationSettings

    ApplicationSettings.objects.update(
        self.application_settings.id,
        nodegraph_zoom=zoom,
        nodegraph_center_x=center_x,
        nodegraph_center_y=center_y,
    )



from airunner.components.application.api.api_service_base import APIServiceBase
from airunner.enums import SignalCode
from airunner.components.nodegraph.data.workflow import Workflow
from airunner.components.nodegraph.gui.widgets.custom_node_graph import (
    CustomNodeGraph,
)


class NodegraphAPIService(APIServiceBase):
    def node_executed(self, node_id, result, data=None):
        self.emit_signal(
            SignalCode.NODE_EXECUTION_COMPLETED_SIGNAL,
            {"node_id": node_id, "result": result, "output_data": data or {}},
        )

    def zoom_changed(self, zoom_level):
        self.emit_signal(SignalCode.NODEGRAPH_ZOOM, {"zoom_level": zoom_level})

    def pan_changed(self, center_x, center_y):
        self.emit_signal(
            SignalCode.NODEGRAPH_PAN,
            {"center_x": center_x, "center_y": center_y},
        )

    def run_workflow(self, graph: CustomNodeGraph):
        self.emit_signal(SignalCode.RUN_WORKFLOW_SIGNAL, {"graph": graph})

    def pause_workflow(self, graph: CustomNodeGraph):
        self.emit_signal(SignalCode.PAUSE_WORKFLOW_SIGNAL, {"graph": graph})

    def stop_workflow(self, graph: CustomNodeGraph):
        self.emit_signal(SignalCode.STOP_WORKFLOW_SIGNAL, {"graph": graph})

    def register_graph(self, graph, nodes_palette, finalize):
        self.emit_signal(
            SignalCode.REGISTER_GRAPH_SIGNAL,
            {
                "graph": graph,
                "nodes_palette": nodes_palette,
                "callback": finalize,
            },
        )

    def load_workflow(self, workflow: Workflow, callback):
        self.emit_signal(
            SignalCode.WORKFLOW_LOAD_SIGNAL,
            {"workflow": workflow, "callback": callback},
        )

    def new_document(self):
        self.clear_workflow(None)

    def clear_workflow(self, callback):
        self.emit_signal(
            SignalCode.CLEAR_WORKFLOW_SIGNAL,
            {"callback": callback},
        )



"""
System and application control tools.

Tools for controlling the application, managing files, and system operations.
"""

import os
from typing import Annotated, Any

from airunner.components.llm.core.tool_registry import tool, ToolCategory


@tool(
    name="quit_application",
    category=ToolCategory.SYSTEM,
    description="Quit the AI Runner application",
    return_direct=True,
    requires_api=True,
    defer_loading=False,  # Essential tool - always available
)
def quit_application(api: Any = None) -> str:
    """Quit the application."""
    api.quit_application()
    return "Quitting application..."


@tool(
    name="toggle_tts",
    category=ToolCategory.SYSTEM,
    description="Enable or disable text-to-speech",
    return_direct=True,
    requires_api=True,
    defer_loading=False,  # Essential tool - always available
)
def toggle_tts(
    enabled: Annotated[bool, "True to enable, False to disable"],
    api: Any = None,
) -> str:
    """Toggle text-to-speech."""
    api.tts.toggle(enabled)
    status = "enabled" if enabled else "disabled"
    return f"Text-to-speech {status}"


@tool(
    name="list_directory",
    category=ToolCategory.FILE,
    description="List files and directories in a path",
    return_direct=False,
    requires_api=False,
)
def list_directory(
    path: Annotated[str, "Directory path to list"],
) -> str:
    """List directory contents."""
    abs_path = os.path.abspath(path)

    if not os.path.exists(abs_path):
        return f"Error: Path does not exist: {abs_path}"

    if not os.path.isdir(abs_path):
        return f"Error: Not a directory: {abs_path}"

    try:
        items = os.listdir(abs_path)
        return "\n".join(sorted(items))
    except PermissionError:
        return f"Error: Permission denied: {abs_path}"
    except Exception as e:
        return f"Error listing directory: {e}"


@tool(
    name="read_file",
    category=ToolCategory.FILE,
    description="Read contents of a text file",
    return_direct=False,
    requires_api=False,
)
def read_file(
    path: Annotated[str, "File path to read"],
) -> str:
    """Read file contents."""
    abs_path = os.path.abspath(path)

    if not os.path.exists(abs_path):
        return f"Error: File does not exist: {abs_path}"

    if not os.path.isfile(abs_path):
        return f"Error: Not a file: {abs_path}"

    try:
        with open(abs_path, "r", encoding="utf-8") as f:
            content = f.read()
        return content
    except PermissionError:
        return f"Error: Permission denied: {abs_path}"
    except UnicodeDecodeError:
        return f"Error: Cannot read file (binary or unsupported encoding): {abs_path}"
    except Exception as e:
        return f"Error reading file: {e}"


@tool(
    name="write_file",
    category=ToolCategory.FILE,
    description="Write content to a text file",
    return_direct=True,
    requires_api=False,
)
def write_file(
    path: Annotated[str, "File path to write to"],
    content: Annotated[str, "Content to write"],
) -> str:
    """Write content to file."""
    abs_path = os.path.abspath(path)

    try:
        # Create parent directories if needed
        os.makedirs(os.path.dirname(abs_path), exist_ok=True)

        with open(abs_path, "w", encoding="utf-8") as f:
            f.write(content)

        return f"Successfully wrote to: {abs_path}"
    except PermissionError:
        return f"Error: Permission denied: {abs_path}"
    except Exception as e:
        return f"Error writing file: {e}"



from typing import Dict

from PySide6.QtCore import Slot
from PySide6.QtGui import QIcon

from airunner.enums import SignalCode
from airunner.components.application.gui.widgets.base_widget import BaseWidget
from airunner.components.home_stage.gui.widgets.templates.knowledge_base_panel_ui import (
    Ui_knowledge_base_panel,
)


class KnowledgeBasePanelWidget(BaseWidget):
    """Small widget for showing document statistics and manual indexing controls.

    This class wraps the generated `Ui_knowledge_base_panel` UI so it can be
    reused in `DocumentsWidget` and other places. It emits the existing
    application signals for RAG indexing actions and listens for progress
    updates.
    """

    widget_class_ = Ui_knowledge_base_panel

    def __init__(self, *args, **kwargs):
        # Track indexing progress locally
        self._total = 0
        self._current = 0
        self._document_name = ""
        self.signal_handlers = {
            SignalCode.RAG_INDEXING_PROGRESS: self.on_indexing_progress,
            SignalCode.RAG_INDEXING_COMPLETE: self.on_indexing_complete,
        }
        super().__init__(*args, **kwargs)

        # Wire up buttons
        try:
            self.ui.index_button.clicked.connect(self._on_index_button_clicked)
        except Exception:
            pass
        try:
            self.ui.cancel_button.clicked.connect(
                self._on_cancel_button_clicked
            )
        except Exception:
            pass

    @Slot()
    def _on_index_button_clicked(self):
        """Emit a RAG index-all signal.

        The existing worker/agent mixins expect SignalCode.RAG_INDEX_ALL_DOCUMENTS
        in order to begin the indexed flow. Provide a simple payload for
        compatibility.
        """
        self.logger.info(
            "KnowledgeBasePanel::Index All clicked - emitting RAG_INDEX_ALL_DOCUMENTS"
        )
        # show infinite progress bar
        self.ui.progress_bar.setRange(0, 0)
        self.emit_signal(SignalCode.RAG_INDEX_ALL_DOCUMENTS, {})

    @Slot()
    def _on_cancel_button_clicked(self):
        """Emit a cancel signal for in-progress indexing flows."""
        self.logger.info(
            "KnowledgeBasePanel::Cancel clicked - emitting RAG_INDEX_CANCEL"
        )
        self.emit_signal(SignalCode.RAG_INDEX_CANCEL, {})

    def on_indexing_progress(self, data: Dict):
        try:
            progress = int(data.get("progress", 0))
            current = int(data.get("current", 0))
            total = int(data.get("total", 0))
            document_name = data.get("documentName", "")
            self._current = current
            self._total = total
            self._document_name = document_name
            if self.ui.progress_bar.maximum() == 0:
                self.ui.progress_bar.setRange(0, 100)
            self.ui.progress_bar.setValue(progress)
        except Exception:
            pass

    def on_indexing_complete(self, data: Dict):
        try:
            self.ui.progress_bar.setValue(100)
            self.ui.progress_text.setText("Complete")
        except Exception:
            pass



"""Expert agents for specialized tasks."""

from airunner.components.agents.expert_agents.calendar_agent import (
    CalendarExpertAgent,
)
from airunner.components.agents.expert_agents.code_agent import (
    CodeExpertAgent,
)
from airunner.components.agents.expert_agents.research_agent import (
    ResearchExpertAgent,
)
from airunner.components.agents.expert_agents.creative_agent import (
    CreativeExpertAgent,
)

__all__ = [
    "CalendarExpertAgent",
    "CodeExpertAgent",
    "ResearchExpertAgent",
    "CreativeExpertAgent",
]



import collections
import re
from enum import Enum

import six

_DEF_PUNCS = ';:,.!?¡¿—…"«»“”'

_PUNC_IDX = collections.namedtuple("_punc_index", ["punc", "position"])


class PuncPosition(Enum):
    """Enum for the punctuations positions"""

    BEGIN = 0
    END = 1
    MIDDLE = 2
    ALONE = 3


class Punctuation:
    """Handle punctuations in text.

    Just strip punctuations from text or strip and restore them later.

    Args:
        puncs (str): The punctuations to be processed. Defaults to `_DEF_PUNCS`.

    Example:
        >>> punc = Punctuation()
        >>> punc.strip("This is. example !")
        'This is example'

        >>> text_striped, punc_map = punc.strip_to_restore("This is. example !")
        >>> ' '.join(text_striped)
        'This is example'

        >>> text_restored = punc.restore(text_striped, punc_map)
        >>> text_restored[0]
        'This is. example !'
    """

    def __init__(self, puncs: str = _DEF_PUNCS):
        self.puncs = puncs

    @staticmethod
    def default_puncs():
        """Return default set of punctuations."""
        return _DEF_PUNCS

    @property
    def puncs(self):
        return self._puncs

    @puncs.setter
    def puncs(self, value):
        if not isinstance(value, six.string_types):
            raise ValueError("[!] Punctuations must be of type str.")
        self._puncs = "".join(
            list(dict.fromkeys(list(value)))
        )  # remove duplicates without changing the oreder
        self.puncs_regular_exp = re.compile(
            rf"(\s*[{re.escape(self._puncs)}]+\s*)+"
        )

    def strip(self, text):
        """Remove all the punctuations by replacing with `space`.

        Args:
            text (str): The text to be processed.

        Example::

            "This is. example !" -> "This is example "
        """
        return re.sub(self.puncs_regular_exp, " ", text).rstrip().lstrip()

    def strip_to_restore(self, text):
        """Remove punctuations from text to restore them later.

        Args:
            text (str): The text to be processed.

        Examples ::

            "This is. example !" -> [["This is", "example"], [".", "!"]]

        """
        text, puncs = self._strip_to_restore(text)
        return text, puncs

    def _strip_to_restore(self, text):
        """Auxiliary method for Punctuation.preserve()"""
        matches = list(re.finditer(self.puncs_regular_exp, text))
        if not matches:
            return [text], []
        # the text is only punctuations
        if len(matches) == 1 and matches[0].group() == text:
            return [], [_PUNC_IDX(text, PuncPosition.ALONE)]
        # build a punctuation map to be used later to restore punctuations
        puncs = []
        for match in matches:
            position = PuncPosition.MIDDLE
            if match == matches[0] and text.startswith(match.group()):
                position = PuncPosition.BEGIN
            elif match == matches[-1] and text.endswith(match.group()):
                position = PuncPosition.END
            puncs.append(_PUNC_IDX(match.group(), position))
        # convert str text to a List[str], each item is separated by a punctuation
        splitted_text = []
        for idx, punc in enumerate(puncs):
            split = text.split(punc.punc)
            prefix, suffix = split[0], punc.punc.join(split[1:])
            splitted_text.append(prefix)
            # if the text does not end with a punctuation, add it to the last item
            if idx == len(puncs) - 1 and len(suffix) > 0:
                splitted_text.append(suffix)
            text = suffix
        while splitted_text[0] == "":
            splitted_text = splitted_text[1:]
        return splitted_text, puncs

    @classmethod
    def restore(cls, text, puncs):
        """Restore punctuation in a text.

        Args:
            text (str): The text to be processed.
            puncs (List[str]): The list of punctuations map to be used for restoring.

        Examples ::

            ['This is', 'example'], ['.', '!'] -> "This is. example!"

        """
        return cls._restore(text, puncs, 0)

    @classmethod
    def _restore(
        cls, text, puncs, num
    ):  # pylint: disable=too-many-return-statements
        """Auxiliary method for Punctuation.restore()"""
        if not puncs:
            return text

        # nothing have been phonemized, returns the puncs alone
        if not text:
            return ["".join(m.punc for m in puncs)]

        current = puncs[0]

        if current.position == PuncPosition.BEGIN:
            return cls._restore(
                [current.punc + text[0]] + text[1:], puncs[1:], num
            )

        if current.position == PuncPosition.END:
            return [text[0] + current.punc] + cls._restore(
                text[1:], puncs[1:], num + 1
            )

        if current.position == PuncPosition.ALONE:
            return [current.mark] + cls._restore(text, puncs[1:], num + 1)

        # POSITION == MIDDLE
        if len(text) == 1:  # pragma: nocover
            # a corner case where the final part of an intermediate
            # mark (I) has not been phonemized
            return cls._restore([text[0] + current.punc], puncs[1:], num)

        return cls._restore(
            [text[0] + current.punc + text[1]] + text[2:], puncs[1:], num
        )



import datetime
from typing import Optional, List
from urllib.parse import urlparse
from sqlalchemy.orm.attributes import flag_modified

from llama_index.core.llms import ChatMessage
from llama_index.core.storage.chat_store.base import BaseChatStore
from llama_index.core.base.llms.types import TextBlock, MessageRole

from airunner.components.llm.data.conversation import Conversation
from airunner.components.llm.utils import strip_names_from_message
from airunner.settings import AIRUNNER_LOG_LEVEL
from airunner.utils.application import get_logger

logger = get_logger(__name__, AIRUNNER_LOG_LEVEL)


class SafeChatMessage(ChatMessage):
    @property
    def content(self) -> str:
        content = ""
        for block in self.blocks:
            if hasattr(block, "text"):
                content += block.text or ""
        return content


class DatabaseChatStore(BaseChatStore):
    def set_messages(self, key: str, messages: list[SafeChatMessage]) -> None:
        """Set messages for a key."""
        index = int(key)
        conversation = Conversation.objects.get(index)

        # TODO: Re-enable this protection after fixing streaming
        # Skip updates if messages contain incomplete assistant responses
        # This prevents race conditions during streaming
        # if messages and len(messages) > 0:
        #     last_message = messages[-1]
        #     if (
        #         hasattr(last_message, "role")
        #         and last_message.role == MessageRole.ASSISTANT
        #         and hasattr(last_message, "blocks")
        #         and len(last_message.blocks) > 0
        #         and hasattr(last_message.blocks[0], "text")
        #         and not last_message.blocks[0].text.strip()
        #     ):
        #         print(
        #             f"Skipping database update for conversation {index} - empty assistant message detected"
        #         )
        #         return

        # Protect against incomplete message overwrites - but allow incremental updates
        # Check if the last message is an incomplete assistant message
        if (
            messages
            and len(messages) > 0
            and hasattr(messages[-1], "role")
            and messages[-1].role == MessageRole.ASSISTANT
            and hasattr(messages[-1], "blocks")
            and messages[-1].blocks
            and hasattr(messages[-1].blocks[0], "text")
        ):
            # For incremental updates, allow updates if:
            # 1. Message has explicit is_complete=False (streaming in progress)
            # 2. Message has is_complete=True (final update)
            # 3. Message text is not empty (has actual content)
            last_message = messages[-1]
            message_dict = (
                last_message.model_dump()
                if hasattr(last_message, "model_dump")
                else last_message
            )

            is_complete = message_dict.get(
                "is_complete", True
            )  # Default to complete if not specified
            has_content = last_message.blocks[0].text.strip()

            # Only skip if message is complete but has no content (likely corrupted)
            if is_complete and not has_content:
                return

        if messages and len(messages) > 0:
            formatted_messages = []
            for message in messages:
                message.blocks[0].text = strip_names_from_message(
                    message.blocks[0].text,
                    conversation.user_name if conversation else None,
                    conversation.chatbot_name if conversation else None,
                )
                formatted_messages.append(message.model_dump())
            messages = formatted_messages

            self.update_or_create(conversation, index, messages, key)

    def update_or_create(self, conversation, index, messages, key):
        if conversation:
            Conversation.objects.update(index, value=messages)
        else:
            conversation = Conversation.objects.create(
                timestamp=datetime.datetime.now(datetime.timezone.utc),
                title="",
                key=key,
                value=messages,
            )
            Conversation.make_current(conversation.id)

    @staticmethod
    def get_latest_chatstore() -> dict:
        """Get the latest chatstore."""
        result = Conversation.objects.order_by("id", "desc").first()
        return (
            {
                "key": str(result.id),
                "value": result.value,
            }
            if result
            else None
        )

    @staticmethod
    def get_chatstores() -> list[dict]:
        """Get all chatstores."""
        result = Conversation.objects.all()
        return [
            {
                "key": str(item.id),
                "value": item.value,
            }
            for item in result
        ]

    def get_messages(self, key: str) -> list[SafeChatMessage]:
        """Get messages for a key. Returns an empty list if key is None or not a valid integer."""
        if key is None or key == "None":
            return []
        try:
            index = int(key)
        except (TypeError, ValueError):
            return []
        result = Conversation.objects.get(index)
        messages = (result.value if result else None) or []
        formatted_messages = []
        for i, message in enumerate(messages):
            role = message.get("role")
            if not isinstance(role, str) or not role:
                role = "user"
            if role == "bot":
                role = "assistant"
            blocks = message.get("blocks")
            sanitized_blocks = []
            if isinstance(blocks, list) and len(blocks) > 0:
                for b in blocks:
                    text = b.get("text") if isinstance(b, dict) else ""
                    if not isinstance(text, str):
                        text = ""
                    sanitized_blocks.append(TextBlock(text=text))
            if not sanitized_blocks:
                sanitized_blocks = [TextBlock(text="")]
            try:
                chat_msg = SafeChatMessage(role=role, blocks=sanitized_blocks)
            except Exception as e:

                logger.warning(
                    f"Failed to construct SafeChatMessage at index {i}: {message} ({e})"
                )
                chat_msg = SafeChatMessage(
                    role=role, blocks=[TextBlock(text="")]
                )
            formatted_messages.append(chat_msg)
        return formatted_messages

    def add_message(self, key: str, message: SafeChatMessage) -> None:
        """Add a message to a conversation. Throws if duplicate."""
        index = int(key)
        conversation = Conversation.objects.get(index)
        messages = (conversation.value or []) if conversation else []
        if message.blocks and len(message.blocks) > 0:
            message.blocks[0].text = message.blocks[0].text.lstrip()
        message.blocks[0].text = strip_names_from_message(
            message.blocks[0].text,
            conversation.user_name if conversation else None,
            conversation.chatbot_name if conversation else None,
        )
        new_msg = message.model_dump()
        messages.append(new_msg)
        self.update_or_create(conversation, index, messages, key)

    def delete_messages(self, key: str) -> Optional[List[SafeChatMessage]]:
        """Delete messages for a key."""
        index = int(key)
        Conversation.objects.delete(index)
        return None

    def delete_message(self, key: str, message_index: int) -> None:
        """Delete a message from a conversation."""
        index = int(key)
        conversation = Conversation.objects.get(index)
        if conversation:
            messages = conversation.value or []
            if 0 <= message_index < len(messages):
                del messages[message_index]
                Conversation.objects.update(index, value=messages)

    def delete_last_message(self, key: str) -> Optional[SafeChatMessage]:
        """Delete last message for a key."""
        index = int(key)
        # First, retrieve the current list of messages
        conversation = Conversation.objects.get(index)
        if conversation:
            messages = conversation.value
        else:
            messages = None

        if messages is None or len(messages) == 0:
            # If the key doesn't exist or the array is empty
            return None

        # Remove the message at the given index
        removed_message = messages[-1]
        messages.pop(-1)
        Conversation.objects.update(index, value=messages)
        return SafeChatMessage.model_validate(removed_message)

    def get_keys(self) -> list[str]:
        """Get all keys."""
        conversations = Conversation.objects.all()
        return [str(conversation.id) for conversation in conversations]

    def update_message(
        self, key: str, message_index: int, new_message: SafeChatMessage
    ) -> None:
        """Update a message in a conversation."""
        try:
            conversation = Conversation.objects.filter_by(current=True)[0]
        except IndexError:
            conversation = None
        if conversation:
            messages = conversation.value or []
            if 0 <= message_index < len(messages):
                new_message.blocks[0].text = strip_names_from_message(
                    new_message.blocks[0].text,
                    conversation.user_name,
                    conversation.chatbot_name,
                )
                messages[message_index] = new_message.model_dump()
                Conversation.objects.update(conversation.id, value=messages)
                flag_modified(conversation, "value")


def params_from_uri(uri: str) -> dict:
    result = urlparse(uri)
    database = result.path[1:]
    return {
        "database": database,
    }



from airunner.components.application.gui.widgets.base_widget import BaseWidget
from airunner.components.stt.gui.widgets.templates.stt_settings_ui import Ui_stt_settings


class STTSettingsWidget(BaseWidget):
    widget_class_ = Ui_stt_settings

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



"""Canvas Position Manager - Single source of truth for all canvas positioning.

This module provides a centralized, tested system for handling canvas positioning
calculations. All position conversions should go through this manager to ensure
consistency and eliminate bugs.

Key Concepts:
- Absolute Position: Position in the canvas coordinate system (persisted to DB)
- Display Position: Position in the viewport (what user sees on screen)
- Grid Origin: The anchor point for grid snapping (center_pos)
- Canvas Offset: User's pan position
- Grid Compensation: Viewport resize adjustment

Formula:
    display_pos = absolute_pos - canvas_offset + grid_compensation
    absolute_pos = display_pos + canvas_offset - grid_compensation
"""

from dataclasses import dataclass
from typing import Tuple, Optional
from PySide6.QtCore import QPointF
import math


@dataclass
class ViewState:
    """Encapsulates the view state needed for position calculations.

    Attributes:
        canvas_offset: User's pan offset (increases when panning right/down)
        grid_compensation: Viewport resize compensation offset
    """

    canvas_offset: QPointF
    grid_compensation: QPointF = None

    def __post_init__(self):
        if self.grid_compensation is None:
            self.grid_compensation = QPointF(0, 0)

    @property
    def total_offset(self) -> QPointF:
        """Combined offset for position calculations."""
        return QPointF(
            self.canvas_offset.x() - self.grid_compensation.x(),
            self.canvas_offset.y() - self.grid_compensation.y(),
        )


class CanvasPositionManager:
    """Manages all canvas position calculations with a single source of truth."""

    @staticmethod
    def absolute_to_display(
        absolute_pos: QPointF, view_state: ViewState
    ) -> QPointF:
        """Convert absolute position to display position.

        Args:
            absolute_pos: Position in canvas coordinates
            view_state: Current view state (offset + compensation)

        Returns:
            Position in viewport coordinates (what user sees)

        Formula:
            display = absolute - canvas_offset + grid_compensation
        """
        return QPointF(
            absolute_pos.x() - view_state.total_offset.x(),
            absolute_pos.y() - view_state.total_offset.y(),
        )

    @staticmethod
    def display_to_absolute(
        display_pos: QPointF, view_state: ViewState
    ) -> QPointF:
        """Convert display position to absolute position.

        Args:
            display_pos: Position in viewport coordinates
            view_state: Current view state (offset + compensation)

        Returns:
            Position in canvas coordinates (for DB persistence)

        Formula:
            absolute = display + canvas_offset - grid_compensation
        """
        return QPointF(
            display_pos.x() + view_state.total_offset.x(),
            display_pos.y() + view_state.total_offset.y(),
        )

    @staticmethod
    def snap_to_grid(
        position: QPointF,
        cell_size: float,
        grid_origin: Optional[QPointF] = None,
        use_floor: bool = True,
        enabled: bool = True,
    ) -> QPointF:
        """Snap a position to grid intersections.

        Args:
            position: Position to snap (in any coordinate system)
            cell_size: Grid cell size in pixels
            grid_origin: Grid anchor point (defaults to 0,0)
            use_floor: If True, snap to top-left. If False, snap to nearest.
            enabled: If False, return position unchanged

        Returns:
            Snapped position in the same coordinate system as input
        """
        if not enabled or cell_size <= 0:
            return QPointF(position)

        # Default grid origin to (0, 0)
        origin_x = grid_origin.x() if grid_origin else 0.0
        origin_y = grid_origin.y() if grid_origin else 0.0

        # Translate to grid-relative coordinates
        rel_x = position.x() - origin_x
        rel_y = position.y() - origin_y

        # Snap in grid space
        if use_floor:
            snapped_rel_x = math.floor(rel_x / cell_size) * cell_size
            snapped_rel_y = math.floor(rel_y / cell_size) * cell_size
        else:
            snapped_rel_x = round(rel_x / cell_size) * cell_size
            snapped_rel_y = round(rel_y / cell_size) * cell_size

        # Translate back to original coordinate system
        return QPointF(
            snapped_rel_x + origin_x,
            snapped_rel_y + origin_y,
        )

    @staticmethod
    def get_centered_position(
        item_size: Tuple[float, float], viewport_size: Tuple[float, float]
    ) -> QPointF:
        """Calculate position to center an item in the viewport.

        Args:
            item_size: (width, height) of item
            viewport_size: (width, height) of viewport

        Returns:
            Absolute position that will center the item
        """
        item_width, item_height = item_size
        viewport_width, viewport_height = viewport_size

        # Calculate position so item center aligns with viewport center
        x = (viewport_width - item_width) / 2.0
        y = (viewport_height - item_height) / 2.0

        return QPointF(x, y)

    @classmethod
    def calculate_drag_position(
        cls,
        initial_absolute_pos: QPointF,
        mouse_delta: QPointF,
        view_state: ViewState,
        cell_size: float = 0,
        grid_origin: Optional[QPointF] = None,
        snap_enabled: bool = False,
    ) -> Tuple[QPointF, QPointF]:
        """Calculate new position during drag operation.

        Args:
            initial_absolute_pos: Item's absolute position when drag started
            mouse_delta: Mouse movement since drag start (in scene coords)
            view_state: Current view state
            cell_size: Grid cell size for snapping
            grid_origin: Grid anchor point
            snap_enabled: Whether to snap to grid

        Returns:
            Tuple of (absolute_position, display_position)
        """
        # Calculate new absolute position
        new_absolute = QPointF(
            initial_absolute_pos.x() + mouse_delta.x(),
            initial_absolute_pos.y() + mouse_delta.y(),
        )

        # Apply grid snapping if enabled
        if snap_enabled and cell_size > 0:
            new_absolute = cls.snap_to_grid(
                new_absolute,
                cell_size,
                grid_origin,
                use_floor=True,
                enabled=True,
            )

        # Convert to display position
        new_display = cls.absolute_to_display(new_absolute, view_state)

        return new_absolute, new_display



import re
from airunner.vendor.openvoice.text.english import (
    english_to_ipa2,
)
from airunner.vendor.openvoice.text.mandarin import (
    chinese_to_ipa,
)


def cjke_cleaners2(text):
    text = re.sub(
        r"\[ZH\](.*?)\[ZH\]", lambda x: chinese_to_ipa(x.group(1)) + " ", text
    )
    text = re.sub(
        r"\[JA\](.*?)\[JA\]",
        lambda x: japanese_to_ipa2(x.group(1)) + " ",
        text,
    )
    text = re.sub(
        r"\[KO\](.*?)\[KO\]", lambda x: korean_to_ipa(x.group(1)) + " ", text
    )
    text = re.sub(
        r"\[EN\](.*?)\[EN\]", lambda x: english_to_ipa2(x.group(1)) + " ", text
    )
    text = re.sub(r"\s+$", "", text)
    text = re.sub(r"([^\.,!\?\-…~])$", r"\1.", text)
    return text



from typing import Any


class SettingsManager:
    """
    Handles application settings logic for MainWindow.
    """

    def __init__(self, app_settings, qsettings, logger=None):
        self.app_settings = app_settings
        self.qsettings = qsettings
        self.logger = logger
        if self.logger:
            self.logger.debug("SettingsManager initialized.")

    def reset_settings(self, main_window):
        if self.logger:
            self.logger.info("Resetting application settings to defaults.")
        self.app_settings.reset_to_defaults()
        if self.logger:
            self.logger.info("Application settings reset to defaults.")
        main_window.restart()

    def update_application_settings(self, key: str, value: Any):
        setattr(self.app_settings, key, value)
        if self.logger:
            self.logger.debug(f"Updated application setting: {key} = {value}")

    def restore_window_state(self, main_window):
        settings = main_window.window_settings
        main_window.setMinimumSize(1024, 1024)
        width = int(settings["width"])
        height = int(settings["height"])
        main_window.resize(width, height)
        x_pos = int(settings["x_pos"])
        y_pos = int(settings["y_pos"])
        main_window.move(x_pos, y_pos)
        if settings.get("is_maximized", False):
            main_window.showMaximized()
        elif settings.get("is_fullscreen", False):
            main_window.showFullScreen()
        else:
            main_window.showNormal()
        main_window.raise_()
        if self.logger:
            self.logger.debug("Restored window state from settings.")
